/**
 * LWSDO, a high performance Domain Driven Design framework.
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * LWSDO is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package org.lwsdo.model;

import org.lwsdo.model.operation.Operation;
import org.lwsdo.util.ObjectCreator;

public class CallInfo {

	private Object           source;
	private Object           target;
	private ExtendedProperty sourceProperty;
	private CallInfo         parent;
	private Settings         settings;	
	private Operation        operation;
	private ObjectCreator    sourceObjectCreator; // Ensure we are not sharing objects between the two object creators 	
	private ObjectCreator    targetObjectCreator; // needed to create the root target data object
	private ProcessingStage  stage;

	public Object getSource() {
		return source;
	}

	public void setSource(Object from) {
		this.source = from;
		validate();
	}

	public Object getTarget() {
		return target;
	}

	public void setTarget(Object to) {
		this.target = to;
		validate();		
	}

	public ExtendedProperty getSourceProperty() {
		return sourceProperty;
	}

	public void setSourceProperty(ExtendedProperty property) {
		this.sourceProperty = property;
	}

	public ExtendedDataObject getParentTargetEntity() {
		CallInfo parent = getParent();
		ExtendedDataObject result = (ExtendedDataObject) parent.getTarget();
		while( ((ExtendedType)result.getType()).isEmbedded() ) {
			parent = parent.getParent();
			result = (ExtendedDataObject) parent.getTarget();
		}

		return result;
	}

	public CallInfo getParent() {
		return parent;
	}

	public void setParent(CallInfo parent) {
		this.parent = parent;
	}

	public Settings getSettings() {
		return settings;
	}

	public void setSettings(Settings settings) {
		this.settings = settings;
	}

	public CallInfo(Object to, CallInfo parent, ExtendedProperty property) {
		this(null, to, parent, property);

		if(!isDataType()) {
			ExtendedDataObject parentSource = (ExtendedDataObject) getParent().getSource();
			ExtendedDataObject sourceRootDO = (ExtendedDataObject) parentSource.getRootObject();

			Object sourceInstance = getSourceFromParent();
			this.source = (sourceInstance == null) ? null : sourceRootDO.getObjectCreator().getExistingTarget(sourceInstance);

			if(sourceInstance != null && source == null) {
				// This is probably not a containment data object so explicitly create one
				this.source = sourceRootDO.getObjectCreator().createDataObject(sourceInstance, property.getType(), parentSource, property);				
			}
		} else
			this.source = getSourceFromParent();

		validate();
	}

	protected void validate() {
		if(!isDataType()) {
			if(this.source != null && !ExtendedDataObject.class.isAssignableFrom(this.source.getClass()))
				throw new IllegalStateException("Source should be a data object but is not treated as one");
			if(this.target != null && !ExtendedDataObject.class.isAssignableFrom(this.target.getClass()))
				throw new IllegalStateException("Target should be a data object but is not treated as one");			
		} else {
			if(this.source != null && ExtendedDataObject.class.isAssignableFrom(this.source.getClass()))
				throw new IllegalStateException("Source should NOT be a data object but is treated as one");
			if(this.target != null && ExtendedDataObject.class.isAssignableFrom(this.target.getClass()))
				throw new IllegalStateException("Target should NOT be a data object but is treated as one");				
		}		
	}

	public CallInfo(Object from, Object to, CallInfo parent, ExtendedProperty property) {

		this.source = from;
		this.target = to;
		this.parent = parent;
		this.sourceProperty = property;

		if(parent != null) {
			operation = parent.getOperation();
			settings = parent.getSettings();
			stage = parent.getStage();
		}
		if(parent != null) {
			this.targetObjectCreator = parent.getTargetObjectCreator();
			this.sourceObjectCreator = parent.getSourceObjectCreator();
		} else
			this.sourceObjectCreator = ((ExtendedDataObject)from).getObjectCreator();

		validate();		
	}

	public boolean isCascadable() {
		if(sourceProperty == null)
			return true;
		return sourceProperty.isContainment();
	}

	public boolean isRequired() {
		return getSourceProperty() != null && !getSourceProperty().isNullable();	
	}

	public boolean isCollection() {
		return getSourceProperty() != null && getSourceProperty().isMany();
	}

	public boolean isDataType() {
		if(getSourceProperty() == null) {
			// This can be null for a collection element, so check the parent
			if(getParent() != null && getParent().getSourceProperty() != null && getParent().getSourceProperty().getElementType() != null)
				return getParent().getSourceProperty().getElementType().isDataType();
			
			if(source != null)
				return !ExtendedDataObject.class.isAssignableFrom(source.getClass());
			if(target != null)
				return !ExtendedDataObject.class.isAssignableFrom(target.getClass());

		} else 
			return getSourceProperty().isDataType();

		return true;
	}

	public String getTargetPropertyPath() {
		/*
		String result = (sourceProperty.getName() != null) ? sourceProperty.getName() : "";
		CallInfo parentInfo = this.parent;
		while(parentInfo != null) {
			result = ((parentInfo.getSourceProperty() != null) ? (parentInfo.getSourceProperty().getName() + Settings.PATH_DELIMITER) : "") + result;
			parentInfo = parentInfo.parent;
		}

		return result;
		*/
		ExtendedDataObject targetDO = (ExtendedDataObject) getTarget();
		return targetDO.getObjectPath();
	}

	public boolean isValid() {
		// TODO: check if the source property is valid to be processed
		// This involves checking the scope, detail and view parameters
		// Also the Cascade level

		return true;
	}

	public Operation getOperation() {
		return operation;
	}

	public void setOperation(Operation operation) {
		this.operation = operation;

		if(targetObjectCreator == null || sourceObjectCreator == null)
			throw new IllegalStateException("The source and target object creators need to be set in order to create the root target object");

		if(target != null)
			throw new IllegalStateException("The root target data object needs to be created before using the operation object");
	}

	public void linkTargetToParent(Object targetValue) {
		ExtendedType targetType = (ExtendedType) ((ExtendedDataObject) getParent().getTarget()).getType();
		ExtendedProperty targetProperty = (ExtendedProperty) targetType.getProperty(getSourceProperty().getName());
		targetProperty.setValue(getParent().getTarget(), targetValue);			
	}

	public Object getTargetFromParent() {
		ExtendedType targetType = (ExtendedType) ((ExtendedDataObject) getParent().getTarget()).getType();
		ExtendedProperty targetProperty = (ExtendedProperty) targetType.getProperty(getSourceProperty().getName());
		return targetProperty.getValue(getParent().getTarget());			
	}

	public ExtendedProperty getTargetProperty() {
		ExtendedType targetType = (ExtendedType) ((ExtendedDataObject) getParent().getTarget()).getType();
		ExtendedProperty targetProperty = (ExtendedProperty) targetType.getProperty(getSourceProperty().getName());

		return targetProperty;
	}

	public Object getSourceFromParent() {
		return ((ExtendedProperty)getSourceProperty()).getValue(getParent().getSource());			
	}

	public ExtendedDataObject getTargetRoot() {
		if(getParent() == null) {
			if(getTarget() == null)
				throw new IllegalStateException("Target root is null, it needs to be created first");

			if(!ExtendedDataObject.class.isAssignableFrom(getTarget().getClass()))
				throw new IllegalStateException("Target root needs to be a data object");

			if ( !((ExtendedDataObject)getTarget()).isRoot() )
				throw new IllegalStateException("Target with null parent needs to be the root data object");

			return (ExtendedDataObject)getTarget();
		} else
			return getParent().getTargetRoot();
	}

	public void setTargetObjectCreator(ObjectCreator oc) {
		this.targetObjectCreator = oc;
	}

	public ObjectCreator getTargetObjectCreator() {
		return this.targetObjectCreator;
	}

	public ProcessingStage getStage() {
		return stage;
	}

	public CallInfo setStage(ProcessingStage stage) {
		this.stage = stage;
		return this;
	}

	public boolean isPathSuffix(String pathSuffix) {
		// Walk through the parent chain to see that the path meets the pathSuffix
		if(getTargetPropertyPath().endsWith(pathSuffix))
			return true;
		else
			return false;
	}

	public boolean isExternal() {
		boolean result = !this.isCollection() && !this.isCascadable();
		for(AssociationSetting setting: settings.getAssociationSettings()) {
			if(setting.isAggregatePart(this)) {
				result = false;
				break;
			}
		}
		return result;
	}

	public ObjectCreator getSourceObjectCreator() {
		return sourceObjectCreator;
	}

	public void setSourceObjectCreator(ObjectCreator sourceObjectCreator) {
		this.sourceObjectCreator = sourceObjectCreator;
	}

	public void clearVisitedTargets() {
		getTargetObjectCreator().clearVisited();
	}

}
