/**
 * LWSDO, a high performance Domain Driven Design framework.
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * LWSDO is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package org.lwsdo.model;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.lwsdo.model.action.CollectionUpdateAction;
import org.lwsdo.model.action.Executable;
import org.lwsdo.model.action.MigratorActionFactory;
import org.lwsdo.model.action.PropertyKey;
import org.lwsdo.model.action.SetterAction;
import org.lwsdo.service.PersistenceOrchestrator;
import org.lwsdo.util.ClassUtil;
import org.lwsdo.util.IdentityArrayList;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

@Component
@Scope(value = "prototype")
public class DependencyGraph {
	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());

	private PersistenceOrchestrator persistenceOrchestrator;
	private Map<Object, Node>       nodeByInstance          = new IdentityHashMap<Object, Node>();
	private IdentityArrayList<Node> transientObjects        = new IdentityArrayList<Node>();
	private Set<Edge>               optionalLinks           = new HashSet<Edge>(); 
	private Set<Edge>               cascadedLinks           = new HashSet<Edge>();
	private Stack<Node>             navigationPath          = new Stack<Node>();	
	private Map<PropertyKey, List<Executable>> propertyActions = new HashMap<PropertyKey, List<Executable>>(); // Includes add, remove and reposition collection elements

	public DependencyGraph(PersistenceOrchestrator persistenceOrchestrator) {
		this.persistenceOrchestrator = persistenceOrchestrator;
	}

	public void clearState() {
		nodeByInstance   = new IdentityHashMap<Object, Node>();
		transientObjects = new IdentityArrayList<Node>();
		optionalLinks    = new HashSet<Edge>();
		cascadedLinks    = new HashSet<Edge>(); 
		navigationPath   = new Stack<Node>();	
	}	

	public Node getNodeByInstance(Object instance) {
		return nodeByInstance.get(instance);
	}

	public static class Edge {
		Node source;         
		Node target;         
		CallInfo callInfo; // property path can be retrieved from this object
		boolean isRequired;
		boolean isCascaded;

		public AbstractProperty getProperty() {
			return (AbstractProperty) callInfo.getSourceProperty();
		}

		public void link() {
			target.addIncomingEdge(this);
			source.addOutgoingEdge(this);	
		}

		public void unlink() {
			target.removeIncomingEdge(this);
			source.removeOutgoingEdge(this);
		}	

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 0;
			result = prime * result + ((source == null) ? 0 : System.identityHashCode(source));
			result = prime * result + ((target == null) ? 0 : System.identityHashCode(target));
			result = prime * result + ((callInfo.getSourceProperty() == null) ? 0 : callInfo.getSourceProperty().getName().hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if(obj == null)
				return false;
			if (this == obj)
				return true;
			if (getClass() != obj.getClass())
				return false;
			Edge other = (Edge) obj;
			String propertyName = (callInfo.getSourceProperty() == null) ? null : callInfo.getSourceProperty().getName();
			String otherPropertyName = (other.callInfo.getSourceProperty() == null) ? null : other.callInfo.getSourceProperty().getName();

			return source == other.source  &&
					target == other.target  &&
					(propertyName == otherPropertyName || (propertyName != null && propertyName.equals(otherPropertyName)));
		}

	}

	public static class CollectionEdge extends Edge {
		/**
		 *  When support for Map and List is added the following properties will be needed 
		 *
		 *  int order;  // for list collections
		 *  Object key; // for maps
		 */
	}

	private void breakOptionalLinks() {
		// It is sufficient to do this for transient objects

		for(Node node: transientObjects) {
			//Node node = nodeByCopy.get(object);
			for(Edge edge: node.getOutgoing()) {			
				if(edge.isRequired == false && edge.isCascaded == false) {
					optionalLinks.add(edge);
					if( isUpdatableProperty(edge) )
						((ExtendedProperty)edge.getProperty()).setValue(edge.source.getInstance(), edge.target.getUnlinkValue()); // clear the link
				}
			}
		}

		for(Edge edge: optionalLinks) {
			edge.unlink();
		}
		logger.debug("Number of broken optional links: " + optionalLinks.size());
	}

	private void breakCascadedLinks() {
		// It is sufficient to do this for transient objects

		for(Node node: transientObjects) {
			//Node node = nodeByCopy.get(object);
			for(Edge edge: node.getOutgoing()) {			
				if(edge.isRequired == false && edge.isCascaded == true) {
					cascadedLinks.add(edge);
					if( isUpdatableProperty(edge) )
						((ExtendedProperty)edge.getProperty()).setValue(edge.source.getInstance(), edge.target.getUnlinkValue()); // clear the link						
				}
			}
		}

		for(Edge edge: cascadedLinks) {
			edge.unlink();
		}
		logger.debug("Number of broken cascaded links: " + cascadedLinks.size());
	}	

	private void removeRequiredOwner(Node nodeToRemove) {
		// If node is a collection, then remove the node from the collection and 
		// delete the object
		// If node is an entity, then remove the cascade owner if it has no required fields
		// pointing to it

		// Find the owner object
		Edge propertyToUnset = null; 
		Node previous = nodeToRemove;
		while(propertyToUnset == null) {
			if(nodeToRemove.getIncoming().size() == 0)
				return; // object has already been deleted
			for(Edge incoming: nodeToRemove.getIncoming()) {
				if(incoming.isCascaded) {
					if(!incoming.isRequired) { 
						propertyToUnset = incoming;
						break;
					} else {
						previous = nodeToRemove;
						nodeToRemove = incoming.source;	// go up to the root by one level
					}
				}
			}
			if(propertyToUnset == null && nodeToRemove == previous) 
				return; // object has already been deleted
		}

		if(Collection.class.isAssignableFrom(propertyToUnset.source.getInstance().getClass())) { // TODO: handle maps
			Object collection = propertyToUnset.source.getInstance();
			Collection persistentCollection = (Collection) collection;
			persistentCollection.remove(nodeToRemove.getInstance());
		} else {
			if(!isUpdatableProperty(propertyToUnset))	
				throw new RuntimeException("Edge should refer to a property for single valued associations");

			((ExtendedProperty)propertyToUnset.getProperty()).setValue(propertyToUnset.source.getInstance(), propertyToUnset.target.getUnlinkValue()); // clear the link			
		}
		propertyToUnset.unlink();	

		// Delete
		persistenceOrchestrator.delete(nodeToRemove.getInstance());
	}

	private void breakNonCascadedRequiredLinks() {
		for(Node node: nodeByInstance.values()) {
			for(Edge edge: node.getOutgoing()) {			
				if(edge.isRequired == true && edge.isCascaded == false && 
						edge.target != null && edge.target.isVisited()) {
					removeRequiredOwner(edge.source);
				}
			}
		}
	}	

	private void getTransientObjects() {
		for(Object instance: nodeByInstance.keySet()) {
			Node node = nodeByInstance.get(instance);

			if(node.isCollectionNode())
				transientObjects.add(node);
			else {
				//if( persistenceOrchestrator.isTransient((ExtendedDataObject)node) )
				Type type = ((ExtendedDataObject)node).getType();
				if(SimpleType.class.isAssignableFrom(type.getClass()))
					continue;
				
				if( !node.isPersistent() && !((ExtendedType)((ExtendedDataObject)node).getType()).isEmbedded() )
					transientObjects.add(node);
			}
		}
	}

	public Edge addEdge(String propertyName, CallInfo ci, Node target) {
		// Only data objects are represented in the graph
		if(ci.isDataType())
			throw new IllegalStateException("Only DataObjects are represented in the graph");

		Edge result = null;
		Node parentNode = (Node) ci.getParent().getTarget();

		Node node = nodeByInstance.get(target.getInstance());
		if(node != null)
			result = node.addIncomingEdge(parentNode, ci);

		return result;
	}

	public void removeNode(ExtendedDataObject copy) {
		Node node = nodeByInstance.get(copy);
		if(node == null)
			return;

		nodeByInstance.remove(copy);
		if(node.getIncoming().size() > 0 || node.getOutgoing().size() > 0)
			throw new RuntimeException("Cannot remove a node that has references pointing to/from it");


	}

	public Node addNode(Node target) {
		if(target.getInstance() == null)
			return null;
		else
			nodeByInstance.put(target.getInstance(), target);

		return target;
	}

	public void clearVisited() {
		for(Node node: nodeByInstance.values()) {
			node.setVisited(false);
		}
	}

	private void ensureIsTree() {
		navigationPath.clear();
		// remove the cycles from the graph and populate the brokenLinks
		for(Node node: nodeByInstance.values()) {
			ensureIsTree(node);
			navigationPath.clear();
		}
	}

	private boolean isUpdatableProperty(Edge edge) {
		return edge.callInfo.getSourceProperty() != null;
	}

	private void setCascadedLinks() {
		// Once the tree is processed, re-link the broken links
		for(Edge edge: cascadedLinks) {
			edge.link();
			if( isUpdatableProperty(edge) )
				((ExtendedProperty)edge.getProperty()).setValue(edge.source.getInstance(), edge.target.getLinkValue()); // set the link
		}
	}

	private void setOptionalLinks() {
		// Once the tree is processed, re-link the broken links
		for(Edge edge: optionalLinks) {
			edge.link();
			if( isUpdatableProperty(edge) )
				((ExtendedProperty)edge.getProperty()).setValue(edge.source.getInstance(), edge.target.getLinkValue()); // set the link						
		}
	}	

	private Node findCascadeRoot(Node node) {
		Node result = node;
		for(Edge edge: node.getIncoming()) {
			if(edge.isCascaded) 
				result = findCascadeRoot(edge.source);
		}

		return result;
	}

	private boolean traverseToRequiredRoot(Node node) {
		boolean foundNewRoot = false;
		for(Edge edge: node.getOutgoing()) {
			if(edge.isRequired) { 
				// either the target is the new root or a new root exists down the tree
				if(transientObjects.contains(edge.target) && !edge.target.isVisited()) { // not a persistent object
					foundNewRoot = true;
					node.setVisited(true);
					logger.debug("Processing non root: " + node.getInstance().getClass().getName());
				}
				if(traverseToRequiredRoot(edge.target)) { // Must recurse
					foundNewRoot = true;
					node.setVisited(true);
					logger.debug("Processing non root: " + node.getInstance().getClass().getName());
				}
			}
		}
		if(!foundNewRoot && node.getInstance() != null) {
			logger.debug("Found required root: " + node.getInstance().getClass().getName());
		}

		return foundNewRoot;
	}	

	private boolean traverseToCascadedRoot(Node node) {
		boolean foundNewRoot = false;
		for(Edge edge: node.getIncoming()) {
			if(edge.isCascaded) { 
				// either the target is the new root or a new root exists down the tree
				if(transientObjects.contains(edge.target) && !edge.target.isVisited()) { // not a persistent object
					foundNewRoot = true;
					node.setVisited(true);
				}
				if(traverseToRequiredRoot(edge.target)) {
					foundNewRoot = true;
					node.setVisited(true);
				}
			}
		}

		return foundNewRoot;
	}		

	private void processRequiredAssociations() {
		clearVisited();
		List<Node> requiredObjects = new ArrayList<Node>();
		for(Node node: transientObjects) {
			//Node node = nodeByCopy.get(object);
			requiredObjects.add(node);
		}

		int level = 0;
		while(requiredObjects.size() > 0) {
			List<Node> nodesToRemove = new ArrayList<Node>();
			logger.debug("********* Processing required objects; level " + level++);
			for(Node node: requiredObjects) {
				traverseToRequiredRoot(node);
			}
			for(Node node: requiredObjects) { // all nodes that have not been visited are the root nodes
				if(node.isVisited() == false) {
					node.setVisited(true);
					nodesToRemove.add(node);
				} else
					node.setVisited(false); // reset it
			}	

			for(Node node: nodesToRemove) {
				if(!node.isCollectionNode())
					persistenceOrchestrator.saveOrUpdate(node.getInstance());
				requiredObjects.remove(node);
				transientObjects.remove(node); // it is no longer a transient object
			}
		}
		// TODO: We are commenting out the flush since we want the flush to be done only after the invariants are checked
		// persistenceOrchestrator.flush();
	}	

	private void processCascadedAssociations() {
		clearVisited();
		List<Node> cascadedObjects = new ArrayList<Node>();
		for(Node node: transientObjects) {
			//Node node = nodeByCopy.get(object);
			cascadedObjects.add(node);
		}

		while(cascadedObjects.size() > 0) {
			List<Node> nodesToRemove = new ArrayList<Node>();
			for(Node node: cascadedObjects) {
				traverseToCascadedRoot(node);
			}
			for(Node node: cascadedObjects) { // all nodes that have not be visited are the root nodes
				if(node.isVisited() == false) {
					node.setVisited(true);
					nodesToRemove.add(node);
				} else
					node.setVisited(false); // reset it
			}			
			for(Node node: nodesToRemove) {
				if(!node.isCollectionNode())				
					persistenceOrchestrator.saveOrUpdate(node.getInstance()); 
				cascadedObjects.remove(node);
				transientObjects.remove(node); // it is no longer a transient object
			}			
		}
		// TODO: We are commenting out the flush since we want the flush to be done only after the invariants are checked
		// persistenceOrchestrator.flush();

		if(transientObjects.size() != 0)
			throw new RuntimeException("Not all transient objects were persisted");
	}

	private void processRemovedCollectionObjects() throws Exception {
		for(Node node: nodeByInstance.values()) {
			if(node.isCollectionNode()) { 
				ExtendedDataObject sourceCollection = (ExtendedDataObject) node.getOwner().callInfo.getSource();
				ExtendedDataObject targetCollection = (ExtendedDataObject) node.getOwner().callInfo.getTarget();

				if(Collection.class.isAssignableFrom(sourceCollection.getInstance().getClass())) { // TODO: handle maps
					Collection source = (Collection) sourceCollection.getInstance();
					node.getOwner().source.removeCollectionObjects(node.getOwner().callInfo, targetCollection, sourceCollection );
				}
			}
		}
	}	

	private void persistTree() {
		clearVisited();

		getTransientObjects();

		// NOTE: once the links are broken, temporarily the copy is out of sync with the key in nodeByCopy, until the links are reattached
		breakOptionalLinks();
		breakCascadedLinks();
		ensureIsTree();

		processRequiredAssociations();
		setCascadedLinks();
		processCascadedAssociations();
		setOptionalLinks();

		try {
			processRemovedCollectionObjects();
			// TODO: invariant processing should occur here
		} catch (Exception e) {
			e.printStackTrace();
		}
	}


	public void visitContainedObjects(ExtendedDataObject root) {
		clearVisited(); // clear the state

		Node rootNode = nodeByInstance.get(root);
		if(rootNode == null)
			throw new RuntimeException("Node for root object cannot be found");

		rootNode.markCascadeTree();
	}	

	public void prepareForDelete(ExtendedDataObject entity) {
		clearVisited();

		Node rootNode = nodeByInstance.get(entity);
		if(rootNode == null)
			throw new RuntimeException("Node for root object cannot be found");		
		rootNode.markCascadeTree();

		// We need to remove all required non-cascaded links to allow the entity delete to work correctly
		breakNonCascadedRequiredLinks();
		persistenceOrchestrator.flush();		
	}

	public void persistGraph(Settings settings) {
		ensureIsTree();
		persistTree();
	}


	private void ensureIsTree(Node node) {

		if(node.isVisited())
			return;

		navigationPath.push(node);


		for(Edge incomingEdge: node.getIncoming()) {
			if(!incomingEdge.isRequired)
				continue;
			Node child = incomingEdge.source;
			if(navigationPath.contains(child)) { // cycle
				throw new RuntimeException("A cycle has been found in the dependency graph");

			}
			ensureIsTree(child);
		}
		node.setVisited(true);

		// We have finished processing this element so it can be popped out
		navigationPath.pop();
	}

	public void processActions(Settings settings) {
		settings.getInterceptor().preActionStage( Collections.unmodifiableMap(propertyActions) );

		for(Map.Entry<PropertyKey, List<Executable>> entry: propertyActions.entrySet()) {
			PropertyKey key = entry.getKey();

			if(key.getProperty().isMany())
				processToMany(key);
			else
				processToOne(key);
		}
	}
	
	public void processToMany(PropertyKey key) {
		for(Executable action: propertyActions.get(key)) {
			action.execute();
		}
	}

	public void processToOne(PropertyKey key) {
		// We follow the system where we execute the action that has a non-null value if one is present or execute the null setting action
		// This policy is present to automatically fix broken links
		// The other policy we can have (maybe through configuration) is to throw an exception if a conflict is detected.
		// A conflict can occur if two actions set different values. It is still desirable to throw an exception if two different non-null values are being set.

		SetterAction actionToExecute = null;
		for(Executable action: propertyActions.get(key)) {
			SetterAction setterAction = (SetterAction) action;
			if(actionToExecute == null)
				actionToExecute = setterAction;
			if( ClassUtil.getInstance(setterAction.getValue()) != null) {
				if(ClassUtil.getInstance(actionToExecute.getValue()) != null && ClassUtil.getInstance(actionToExecute.getValue()) != ClassUtil.getInstance(setterAction.getValue()) )
					throw new IllegalArgumentException("Two different objects cannot share a reference to the same object in a ToOne bi-directional relationship.");
				else
					actionToExecute = setterAction;

			}
		}
		actionToExecute.execute();		
	}

	public void addAction(Executable action) {

		List<Executable> actions = propertyActions.get(action.getKey());
		if(actions == null) {
			actions = new ArrayList<Executable>();
			propertyActions.put((PropertyKey) action.getKey(), actions);
		}
		actions.add(action);
	}

	public CollectionUpdateAction getOrCreateMigratorAction(PropertyKey key) {
		// actions should not be null
		List<Executable> actions = propertyActions.get(key);

		CollectionUpdateAction migrator = null;		
		if(actions == null) {
			migrator = MigratorActionFactory.getInstance(key);
			addAction(migrator);
			return migrator;
		}

		if(actions.size() == 1) {
			migrator = (CollectionUpdateAction) actions.get(0);
		} else if(actions.size() == 0 || actions.size() > 1)
			throw new IllegalStateException("Collection property should have only one migrator action");

		return migrator;
	}

}