/**
 * LWSDO, a high performance Domain Driven Design framework.
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * LWSDO is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package org.lwsdo.model;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import javax.persistence.AccessType;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.lwsdo.service.DataAccessService;

public class DerivedType extends AbstractType {
	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());	

	private ExtendedType                 referenceType;
	private Class<?>                     javaClass;
	private List<Type>                   baseTypes;	

	public DerivedType(ExtendedType referenceType, Class<?> javaClass) {
		this.referenceType = referenceType;
		this.javaClass = javaClass;
		
		setImmutable(referenceType.isImmutable());
	}

	public void setProperty(DataAccessService dataAccessService) {
		if(properties == null) {
			// populate the properties for this type
			properties = new HashMap<String, Property>();	
			for(Property referenceProperty: referenceType.getProperties()) {
				AbstractProperty abstractProperty = (AbstractProperty) referenceProperty;
				Type propertyType = dataAccessService.getDerivedType(dataAccessService.getTypeMapper().toDerived(referenceProperty.getType().getInstanceClass()));
				DerivedProperty derivedProperty = new DerivedProperty((ExtendedProperty) referenceProperty, propertyType, this);
				derivedProperty.init(dataAccessService);
				logger.debug("[" + getName() + "] reference property name: " + abstractProperty.getName() + ", type name: " + derivedProperty.getJavaType());
				properties.put(derivedProperty.getName(), derivedProperty);
			}			
		} 		
	}	
	
	public void setOpposite(DataAccessService das) {
		for(Property property: properties.values())
			((DerivedProperty)property).initMappedBy(das);
	}

	@Override
	public String getName() {
		return javaClass.getName();
	}
	
	@Override
	public String getEntityName() {
		return referenceType.getName();
	}	

	@Override
	public String getURI() {
		return null;
	}

	@Override
	public Class<?> getInstanceClass() {
		return this.javaClass;
	}

	@Override
	public boolean isInstance(Object object) {
		return getInstanceClass().isAssignableFrom(object.getClass());
	}

	@Override
	public boolean isDataType() {
		return referenceType.isDataType();
	}

	@Override
	public boolean isOpen() {
		return false;
	}

	@Override
	public boolean isSequenced() {
		return false;
	}

	@Override
	public boolean isAbstract() {
		return false;
	}

	@Override
	public List<Type> getBaseTypes() {
		return baseTypes;
	}

	public void setBaseType(DataAccessService dataAccessService) {
		baseTypes = new ArrayList<Type>();

		for(Type baseType: referenceType.getBaseTypes()) {
			Class<?> derivedClass = getTypeMapper().toDerived(baseType.getInstanceClass());
			Type derivedType = dataAccessService.getType(derivedClass);

			baseTypes.add(derivedType);
		}
	}

	@Override
	public List<Property> getDeclaredProperties() {
		List<Property> result = new ArrayList<Property>();
		for(Property property: referenceType.getDeclaredProperties()) {
			logger.debug("[" + getName() + "] declared reference property name: " + property.getName());
			result.add(getProperty(property.getName()));
		}

		return result;
	}

	@Override
	public List<?> getAliasNames() {
		return new ArrayList<String>();
	}

	@Override
	public List<?> getInstanceProperties() {
		return new ArrayList<Object>();
	}

	@Override
	public Object get(Property property) {
		return null;
	}

	@Override
	public AccessType getAccessType() {
		// Prefer property access for derived types
		return AccessType.PROPERTY;
	}

	@Override
	public boolean isReferenceType() {
		return false;
	}

	@Override
	public ExtendedType getReferenceType() {
		return referenceType;
	}

	@Override
	public Property getIdentifierProperty() {
		return getProperty(referenceType.getIdentifierProperty().getName());
	}

	@Override
	public boolean isEmbedded() {
		return referenceType.isEmbedded();
	}

	@Override
	public Property getVersionProperty() {
		return getProperty(referenceType.getVersionProperty().getName());
	}

	@Override
	public Property getUserKey() {
		return getProperty(referenceType.getUserKey().getName());
	}

	@Override
	public Property getCollectionUserKey() {
		return getProperty(referenceType.getCollectionUserKey().getName());
	}

}
