/**
 * LWSDO, a high performance Domain Driven Design framework.
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * LWSDO is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package org.lwsdo.model;

import java.util.ArrayList;
import java.util.List;

import javax.persistence.metamodel.Attribute.PersistentAttributeType;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.hibernate.engine.spi.CascadeStyle;
import org.hibernate.engine.spi.CascadingAction;
import org.hibernate.engine.spi.SessionFactoryImplementor;
import org.hibernate.mapping.ManyToOne;
import org.hibernate.mapping.OneToOne;
import org.hibernate.mapping.Value;
import org.hibernate.persister.collection.CollectionPersister;
import org.hibernate.type.CollectionType;
import org.lwsdo.service.DataAccessService;
import org.lwsdo.service.HibernateDAS;
import org.lwsdo.util.ClassUtil;

public class HibernateProperty extends AbstractProperty {
	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());	

	private org.hibernate.mapping.Property hibernateProperty;
	private Type                           keyType;  // If an index is present the property represents a List, else it represents a bag. 
                                                     // Just because this flag is true does not mean listIndexProperty needs to be present, since the index could be maintained
                                                     // in a JOIN table
	private Type                           elementType;
	// private String                         listIndexProperty; // If the index is mapped on the entity using the ListIndex annotation

	public HibernateProperty(org.hibernate.mapping.Property hibernateProperty, Type type, HibernateType parentType) {
		super(type, parentType);
		this.hibernateProperty = hibernateProperty;

		init();		
	}	

	@Override
	public String getName() {
		return hibernateProperty.getName();
	}

	@Override
	public boolean isMany() {
		// TODO: handle CollectionType and MapType
		return hibernateProperty.getType().isCollectionType();
	}

	@Override
	public boolean isContainment() {
		CascadeStyle style = hibernateProperty.getCascadeStyle();
		return (style.doCascade(CascadingAction.SAVE_UPDATE) || style.doCascade(CascadingAction.PERSIST)) || 	
				hibernateProperty.getType().isComponentType() || super.isContainment();
	}

	/**
	 * Not directly supported in Hibernate
	 */
	@Override
	public Object getDefault() {
		return null;
	}

	@Override
	public boolean isReadOnly() {
		return !hibernateProperty.isUpdateable();
	}

	@Override
	public List<?> getAliasNames() {
		return new ArrayList<String>();
	}

	@Override
	public boolean isNullable() {
		return hibernateProperty.isOptional();
	}

	@Override
	public boolean isOpenContent() {
		return false;
	}

	@Override
	public List<?> getInstanceProperties() {
		return new ArrayList<Object>();
	}

	@Override
	public Object get(Property property) {
		return null;
	}		

	@Override
	public void init(DataAccessService das) {
		HibernateDAS hibernateDAS = (HibernateDAS) das;

		if(isMany()) {
			if(hibernateProperty.getType().isCollectionType()) {
				CollectionType collType = (CollectionType) hibernateProperty.getType();
				CollectionPersister cp = ((SessionFactoryImplementor) hibernateDAS.getSessionFactory()).getCollectionPersister( collType.getRole() );
				keyType = (cp.getIndexType() == null) ? null : das.getType(cp.getIndexType().getReturnedClass());
				
				elementType = das.getType(cp.getElementType().getReturnedClass());
			}
		} 
	}

	@Override
	public Type getKeyType() {
		return keyType;
	}

	@Override
	public Type getElementType() {
		return elementType;
	}

	@Override
	public void initMappedBy(DataAccessService das) {
		String mappedBy = ClassUtil.getMappedBy(getContainingType().getInstanceClass(), getName());

		Type type = isMany() ? getElementType() : getType();
		setMappedBy( mappedBy != null ? type.getProperty(mappedBy) : null, mappedBy);

		if(getMappedBy() != null)
			logger.debug("Opposite of property '" + getContainingType().getName() + "." + getName() + "' is '" + mappedBy + "'");
	}

	@Override
	public PersistentAttributeType getAssociationType() {
		Value value = hibernateProperty.getValue();

		if(org.hibernate.mapping.Collection.class.isAssignableFrom(value.getClass())) {
			org.hibernate.mapping.Collection collectionValue = (org.hibernate.mapping.Collection) value;
			Value element = collectionValue.getElement();

			if(ManyToOne.class.isAssignableFrom(element.getClass()))
				return PersistentAttributeType.MANY_TO_MANY;
			else
				return PersistentAttributeType.ONE_TO_MANY;
		}

		if(OneToOne.class.isAssignableFrom(value.getClass()))
			return PersistentAttributeType.ONE_TO_ONE;		

		if(ManyToOne.class.isAssignableFrom(value.getClass())) {
			if( ((ManyToOne)value).isLogicalOneToOne() )
				return PersistentAttributeType.ONE_TO_ONE;
			else
				return PersistentAttributeType.MANY_TO_ONE;
		}

		return null;		
	}

	@Override
	public boolean isMap() {
		if(isMany())
		{
			Value value = hibernateProperty.getValue();
			if(org.hibernate.mapping.Map.class.isAssignableFrom(value.getClass()))
				return true;
		}

		return false;
	}

	@Override
	public boolean isList() {
		if(isMany() && getKeyType() != null)
		{
			Value value = hibernateProperty.getValue();
			if(org.hibernate.mapping.List.class.isAssignableFrom(value.getClass()))
				return true;
		}

		return false;
	}

	@Override
	public boolean isSet() {
		if(isMany())
		{
			Value value = hibernateProperty.getValue();
			if(org.hibernate.mapping.Set.class.isAssignableFrom(value.getClass()))
				return true;
		}

		return false;
	}

}
