/**
 * LWSDO, a high performance Domain Driven Design framework.
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * LWSDO is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package org.lwsdo.model;

import java.util.ArrayList;
import java.util.List;

import javax.persistence.AccessType;
import javax.persistence.metamodel.Attribute.PersistentAttributeType;

import org.hibernate.engine.spi.CascadeStyle;
import org.hibernate.engine.spi.CascadingAction;
import org.hibernate.metamodel.binding.AssociationAttributeBinding;
import org.hibernate.metamodel.binding.AttributeBinding;
import org.hibernate.metamodel.binding.ManyToOneAttributeBinding;
import org.hibernate.metamodel.binding.SingularAssociationAttributeBinding;
import org.hibernate.metamodel.binding.SingularAttributeBinding;
import org.hibernate.metamodel.domain.Attribute;
import org.hibernate.metamodel.domain.IndexedPluralAttribute;
import org.hibernate.metamodel.domain.PluralAttribute;
import org.hibernate.metamodel.domain.PluralAttributeNature;
import org.hibernate.metamodel.domain.SingularAttribute;
import org.lwsdo.service.DataAccessService;
import org.lwsdo.service.JPADataAccessService;
import org.lwsdo.service.MetadataDAS;
import org.lwsdo.util.ClassUtil;

public class MetadataProperty extends AbstractProperty {

	private Attribute     attribute;
	private boolean       readOnly;	
	private Type          keyType;
	private Type          elementType;		

	public MetadataProperty(Attribute attribute, Type type, MetadataType parentType) {
		super(type, parentType);
		this.attribute = attribute;
	}	

	@Override
	public String getName() {
		return attribute.getName();
	}

	public static Class<?> getClass(Attribute attribute) {
		if(PluralAttribute.class.isAssignableFrom(attribute.getClass())) {
			PluralAttribute pluralAttribute = (PluralAttribute) attribute;
			return pluralAttribute.getElementType().getClassReference();
		} else if(SingularAttribute.class.isAssignableFrom(attribute.getClass())) {
			SingularAttribute singularAttribute = (SingularAttribute) attribute;
			return singularAttribute.getSingularAttributeType().getClassReference();	
		}

		return null;
	}

	@Override
	public boolean isMany() {
		return !attribute.isSingular();
	}

	@Override
	public boolean isContainment() {
		boolean result = false;

		AttributeBinding attributeBinding = ((MetadataType)getContainingType()).getEntityBinding().locateAttributeBinding(attribute.getName());
		if(AssociationAttributeBinding.class.isAssignableFrom(attributeBinding.getClass())) {
			AssociationAttributeBinding associationAttrBinding = (AssociationAttributeBinding) attributeBinding;
			CascadeStyle style = associationAttrBinding.getCascadeStyle();

			result = style.doCascade(CascadingAction.SAVE_UPDATE) || style.doCascade(CascadingAction.PERSIST);		
		}

		return result || super.isContainment();
	}

	/**
	 * Not directly supported in Hibernate
	 */
	@Override
	public Object getDefault() {
		return null;
	}

	@Override
	public boolean isReadOnly() {
		return readOnly;	
	}
	
	public void setReadOnly(JPADataAccessService das) {
		readOnly = !ClassUtil.isUpdatable(((MetadataType)getContainingType()).getInstanceClass(), getName());	
	}	

	/**
	 * Valid for bi-directional links
	 * @param das
	 */
	@Override
	public void initMappedBy(DataAccessService das) {
		String mappedBy = null;

		AttributeBinding attributeBinding = ((MetadataType)getContainingType()).getEntityBinding().locateAttributeBinding(attribute.getName());
		if(SingularAssociationAttributeBinding.class.isAssignableFrom(attributeBinding.getClass())) {
			SingularAssociationAttributeBinding singularBinding = (SingularAssociationAttributeBinding) attributeBinding;
			mappedBy = singularBinding.getReferencedAttributeName();
		}
		if(mappedBy == null)
			mappedBy = ((MetadataDAS)das).getMappings().getFromMappedBy(getType().getName(), getName());

		setMappedBy(mappedBy != null ? getType().getProperty(mappedBy) : null, mappedBy);
	}

	@Override
	public List<?> getAliasNames() {
		return new ArrayList<String>();
	}

	@Override
	public boolean isNullable() {
		boolean result = true;

		AttributeBinding attributeBinding = ((MetadataType)getContainingType()).getEntityBinding().locateAttributeBinding(attribute.getName());		
		if(SingularAttributeBinding.class.isAssignableFrom(attributeBinding.getClass())) {
			SingularAttributeBinding singularAttrBinding = (SingularAttributeBinding) attributeBinding;
			result = singularAttrBinding.isNullable();
		}

		return result;
	}

	@Override
	public boolean isOpenContent() {
		return false;
	}

	@Override
	public List<?> getInstanceProperties() {
		return new ArrayList<Object>();
	}

	@Override
	public Object get(Property property) {
		return null;
	}
	
	@Override
	public AccessType getAccessType() {
		AttributeBinding attributeBinding = ((MetadataType)getContainingType()).getEntityBinding().locateAttributeBinding(attribute.getName());	
		String accessType = attributeBinding.getPropertyAccessorName();
		return ClassUtil.getAccessStrategy(accessType);
	}

	@Override
	public Type getKeyType() {
		return keyType;
	}

	@Override
	public Type getElementType() {
		return elementType;
	}

	@Override
	public void init(DataAccessService das) {
		if( PluralAttribute.class.isAssignableFrom(attribute.getClass()) ) {
			PluralAttribute pluralAttribute = (PluralAttribute) attribute;
			elementType = das.getType(pluralAttribute.getElementType().getClassReference());
			if(IndexedPluralAttribute.class.isAssignableFrom(attribute.getClass())) {
				IndexedPluralAttribute mapAttribute = (IndexedPluralAttribute) attribute;
				keyType = das.getType(mapAttribute.getIndexType().getClassReference());
			}
		}
	}

	@Override
	public PersistentAttributeType getAssociationType() {	
		
		if( SingularAssociationAttributeBinding.class.isAssignableFrom(attribute.getClass()) ) {
			if(ManyToOneAttributeBinding.class.isAssignableFrom(attribute.getClass()))
				return PersistentAttributeType.MANY_TO_ONE;
			else
				return PersistentAttributeType.ONE_TO_ONE;
		} else if( PluralAttribute.class.isAssignableFrom(attribute.getClass()) ) {
			// TODO: How do you identify MANY_TO_MANY?
			return PersistentAttributeType.ONE_TO_MANY;
		}
		
		return null;
	}

	@Override
	public boolean isMap() {
		if(isMany()) {
			if( PluralAttribute.class.isAssignableFrom(attribute.getClass()) ) {
				PluralAttribute pluralAttribute = (PluralAttribute) attribute;
				if(pluralAttribute.getNature() == PluralAttributeNature.MAP)
					return true;
			}
		}
		
		return false;
	}

	@Override
	public boolean isList() {
		if(isMany() && getKeyType() != null) {
			if( PluralAttribute.class.isAssignableFrom(attribute.getClass()) ) {
				PluralAttribute pluralAttribute = (PluralAttribute) attribute;
				if(pluralAttribute.getNature() == PluralAttributeNature.LIST)
					return true;
			}
		}
		
		return false;

	}

	@Override
	public boolean isSet() {
		if(isMany()) {
			if( PluralAttribute.class.isAssignableFrom(attribute.getClass()) ) {
				PluralAttribute pluralAttribute = (PluralAttribute) attribute;
				if(pluralAttribute.getNature() == PluralAttributeNature.SET)
					return true;
			}
		}
		
		return false;
	}	

}
