/**
 * LWSDO, a high performance Domain Driven Design framework.
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * LWSDO is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package org.lwsdo.model.action;

import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;

import org.lwsdo.model.AggregateAction;
import org.lwsdo.model.CallInfo;
import org.lwsdo.model.ExtendedDataObject;
import org.lwsdo.model.ExtendedProperty;
import org.lwsdo.model.event.CollectionElement;
import org.lwsdo.model.event.CollectionEvent;
import org.lwsdo.util.ClassUtil;


public abstract class CollectionUpdateAction implements Executable {

	public abstract void addAction(Executable action);

	public abstract void addOrReposition(CallInfo callInfo) throws Exception;
	
	@Override
	public Executable getTriggeringAction() {
		return null;
	}

	public void removeObsoletes(CallInfo callInfo) {

		/* algorithm
		 * 1. All the objects in the targetCollection should have ids since these are entities
		 * 2. We get a list of all ids from the sourceCollection
		 * 3. All ids that are not present in the sourceCollection are obsolete and needs to be removed
		 * 
		 * For a Map, we only honor entity in the value and not in the key. So we just need to check the values collection
		 * of the source and target maps.
		 */		

		Map targetKeys = new HashMap<Object, Object>();
		for(ExtendedDataObject element: ((ExtendedDataObject)callInfo.getTarget()).getList(callInfo))
			targetKeys.put(element.getIdentifierValue(), element);		
		
		Map sourceKeys = new HashMap<Object, Object>();		
		if(callInfo.getSource() != null) 
			for(ExtendedDataObject element: ((ExtendedDataObject)callInfo.getSource()).getList(callInfo)) {
				Object idValue = element.getIdentifierValue();
				if(idValue != null)
					sourceKeys.put(idValue, element);
			}			

		for(Object entry : targetKeys.entrySet()) {
			Object targetKey = ((Map.Entry) entry).getKey();
			Object targetElement = ((Map.Entry) entry).getValue();

			if(!sourceKeys.containsKey(targetKey)) { // obsolete element being marked for removal
				CallInfo next = new CallInfo(null, null, callInfo, null);
				next.setTarget(callInfo.getTargetObjectCreator().createTarget(next, targetElement, null));
				removeCollectionElement(next, (ExtendedDataObject) next.getTarget());
			}
		}		
	}

	protected void removeCollectionElement(CallInfo callInfo, ExtendedDataObject collectionElement) {

		// TODO: List and Map might need to add additional info to the RemoveElementAction. So this method will need to be overridden by the subclasses.
		PropertyKey propertyKey = new PropertyKey((ExtendedDataObject) callInfo.getParent().getTarget(), callInfo.getParent().getTargetProperty());
		Executable originalAction = new RemoveElementAction(propertyKey, collectionElement, callInfo, null, null);
		this.addAction(originalAction);		

		// Set the backRef value to null. This is an asynchronous activity
		PropertyKey newOppositeKey = new PropertyKey(collectionElement, callInfo.getParent().getTargetProperty().getOpposite());			
		callInfo.getTargetRoot().getDependencyGraph().addAction(new SetterAction(null, newOppositeKey, originalAction)); 	

	}		
}
