/**
 * LWSDO, a high performance Domain Driven Design framework.
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * LWSDO is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package org.lwsdo.model.action;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.lwsdo.model.AggregateAction;
import org.lwsdo.model.CallInfo;
import org.lwsdo.model.ExtendedDataObject;
import org.lwsdo.model.ExtendedProperty;
import org.lwsdo.model.event.CollectionElement;
import org.lwsdo.model.event.CollectionEvent;
import org.lwsdo.model.operation.AbstractOperation;
import org.lwsdo.util.ClassUtil;


public class SetUpdateAction extends CollectionUpdateAction {
	
	private PropertyKey key;
	private List<AddElementAction> addActions = new ArrayList<AddElementAction>();
	private List<RemoveElementAction> removeActions = new ArrayList<RemoveElementAction>(); 

	public SetUpdateAction(PropertyKey key) {
		this.key = key;
	}
	
	@Override
	public void execute() {
	
		// Remove obsolete elements
		for(RemoveElementAction action: removeActions) {
			// Remove the element from the set
			customRemoveFromSet(action.getCallInfo());
		}
		
		// Add new elements
		for(AddElementAction action: addActions) {
			// Remove the element from the set
			customAddToSet(action.getCallInfo());
		}		
	}
	
	protected void customAddToSet(CallInfo callInfo) {
		if(callInfo.getSettings().getAction() == AggregateAction.VERIFY) // Do not modify if this is just verification
			return;

		Object collection                   = (java.util.Collection) callInfo.getParent().getTargetFromParent();
		Object collectionElement            = callInfo.getTarget();
		ExtendedProperty collectionProperty = callInfo.getParent().getTargetProperty();

		// check if there is a custom collection add
		Method customCollectionAdd = ClassUtil.getCustomAdd(collectionProperty.getElementType().getInstanceClass(), collectionProperty.getName());
		if(customCollectionAdd != null) {
			CollectionEvent event = new CollectionElement(callInfo, collection, collectionElement);
			collectionProperty.setCustomAdd(collection, event);
		} else
			((java.util.Collection<Object>)collection).add(ClassUtil.getInstance(collectionElement));		
	}	

	protected void customRemoveFromSet(CallInfo callInfo) {
		if(callInfo.getSettings().getAction() == AggregateAction.VERIFY) // Do not modify if this is just verification
			return;

		Object collection                   = (java.util.Collection) callInfo.getParent().getTargetFromParent();
		Object collectionElement            = callInfo.getTarget();
		ExtendedProperty collectionProperty = callInfo.getParent().getTargetProperty();

		// check if there is a custom collection add
		Method customCollectionRemove = ClassUtil.getCustomRemove(collectionProperty.getElementType().getInstanceClass(), collectionProperty.getName());
		if(customCollectionRemove != null) {
			CollectionEvent event = new CollectionElement(callInfo, collection, collectionElement);
			collectionProperty.setCustomRemove(collection, event);
		} else
			((java.util.Collection<Object>)collection).remove(ClassUtil.getInstance(collectionElement));		
	}		

	@Override
	public void addAction(Executable action) {
		if(AddElementAction.class.isAssignableFrom(action.getClass()))
			addActions.add((AddElementAction) action);
		else if(RemoveElementAction.class.isAssignableFrom(action.getClass()))
			removeActions.add((RemoveElementAction) action);
		else
			throw new IllegalArgumentException("A set can have only add and remove actions");
	}
	
	public void addOrReposition(CallInfo callInfo) throws Exception {

		Map<Object, ExtendedDataObject> targetKeys = new HashMap<Object, ExtendedDataObject>();
		for(ExtendedDataObject element: ((ExtendedDataObject)callInfo.getTarget()).getList(callInfo))
			targetKeys.put(element.getIdentifierValue(), element);		
			
		if(callInfo.getSource() != null) {
			AbstractOperation oper = (AbstractOperation) callInfo.getOperation();	
			for(ExtendedDataObject childDO: ((ExtendedDataObject)callInfo.getSource()).getList(callInfo)) {
				CallInfo next = new CallInfo(childDO, null, callInfo, null);
	
				ExtendedDataObject targetElement = targetKeys.get(childDO.getIdentifierValue());
				if(callInfo.getSettings().getAction() != AggregateAction.CREATE && targetElement != null) { // updating an existing object

					targetElement.addNode();

					next.setTarget(targetElement);
					oper.processValue(next);

				} else {

					next.setTarget(oper.createTarget(next, null));
					oper.processValue(next);

					// AddElementAction
					addSetElement(next, (ExtendedDataObject) next.getTarget());
				}
			}	
		}

	}
	

	protected void addSetElement(CallInfo callInfo, ExtendedDataObject collectionElement) {

		PropertyKey propertyKey = new PropertyKey((ExtendedDataObject) callInfo.getParent().getTarget(), callInfo.getParent().getTargetProperty());
		Executable originalAction = new AddElementAction(propertyKey, collectionElement, callInfo, null, null);
		this.addAction(originalAction);		

		// Set the backRef of value to point to the parent target (or to the ancester in case of embedded property)
		// TODO: handle MANY_TO_MANY backRef
		PropertyKey newOppositeKey = new PropertyKey(collectionElement, callInfo.getParent().getTargetProperty().getOpposite());			
		callInfo.getTargetRoot().getDependencyGraph().addAction(new SetterAction(collectionElement, newOppositeKey, originalAction)); 				
	}		
	

	@Override
	public PropertyKey getKey() {
		return key;
	}

}
