/**
 * LWSDO, a high performance Domain Driven Design framework.
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * LWSDO is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package org.lwsdo.model.operation;

import java.lang.reflect.Method;
import java.util.Map;

import org.lwsdo.exception.PropertyMappingException;
import org.lwsdo.model.AggregateAction;
import org.lwsdo.model.CallInfo;
import org.lwsdo.model.ContentDetail;
import org.lwsdo.model.ExtendedDataObject;
import org.lwsdo.model.ExtendedProperty;
import org.lwsdo.model.ExtendedType;
import org.lwsdo.model.Node;
import org.lwsdo.model.ProcessingStage;
import org.lwsdo.model.Property;
import org.lwsdo.model.event.CollectionElement;
import org.lwsdo.model.event.CollectionEvent;
import org.lwsdo.util.ClassUtil;

public abstract class AbstractOperation implements Operation {		

	public void execute(CallInfo callInfo) {
		// Create new objects
		this.process(callInfo.setStage(ProcessingStage.CREATE)); 
		callInfo.clearVisitedTargets();		

		// TODO: get list of revisions for this aggregate and iterate this process for each revision
		// The upgrade and downgrade methods will be called for revision processing instead of the process method
		// Step 1: Create the update actions
		((Node) callInfo.getTarget()).setVisited(true);
		this.process(callInfo.setStage(ProcessingStage.UPDATE));
		callInfo.clearVisitedTargets();		
		// Step 2: Execute the actions
		callInfo.getTargetRoot().getDependencyGraph().processActions(callInfo.getSettings());

		// Process invariants 
		this.process(callInfo.setStage(ProcessingStage.INVARIANT));
		callInfo.clearVisitedTargets();	
		
		setResult(callInfo.getTarget());
	}

	protected void setResult(Object target) {}

	protected boolean isAllowed(CallInfo callInfo, ExtendedDataObject source, Property sourceProperty) {
		return (callInfo.getStage() == ProcessingStage.CREATE &&  (((ExtendedType)source.getType()).isImmutable() || 
				sourceProperty.isContainment()                  ||
				sourceProperty.getType().isDataType()))         
				||
				callInfo.getStage() != ProcessingStage.CREATE;

	}

	protected void processInvariants(CallInfo callInfo) {
		if(callInfo.getStage() == ProcessingStage.INVARIANT) {
			ExtendedDataObject target = (ExtendedDataObject) callInfo.getTarget();
			if(callInfo.isCascadable())
				target.checkInvariants();
		}		
	}

	/* Algorithm
	 * =========
	 * 
	 * 1. Process the "from" data object
	 *    a) If it has properties then process each on in turn
	 *    b) If it does not have properties and is a collection/map then process each element in turn
	 *    c) If it does not have properties then process the data type
	 * 2. If it is a valid property, call ObjectCreator#toDataObject on the "from" property and goto STEP 1
	 * 3. Else set as null in the "to" object and process the next item
	 * 
	 */		
	public void process(CallInfo callInfo) {

		try {
			if(!callInfo.isDataType()) {
				ExtendedDataObject source = (ExtendedDataObject) callInfo.getSource();
				for(Property sourceProperty: source.getType().getProperties()) {
					if(isAllowed(callInfo, source, sourceProperty)) {
						CallInfo next = new CallInfo(null, callInfo, (ExtendedProperty) sourceProperty);
						processValue(next); // recurse
					}
				}

				processInvariants(callInfo);
			}

		} catch (Exception e) {
			throw ClassUtil.wrapRun(e);
		} 
	}

	private void processToMany(CallInfo callInfo) throws Exception {
		if(callInfo.getTarget() == null)
			return;

		processCollection(callInfo);
	}

	protected void setModified(Object dataObject) {
		ExtendedDataObject obj = (ExtendedDataObject) dataObject;
		while( ((ExtendedType)obj.getType()).isEmbedded() )
			obj = (ExtendedDataObject) obj.getContainer();
		((Node)obj).setModified(true);
	}

	public void processValue(CallInfo ci) throws Exception {

		ci.getOperation().preCreatePropertyCallback(ci);

		if(ci.getSource() == null) {
			if ( ci.getTargetFromParent() != null && shouldUnlink(ci)) {
				processNullValue(ci);
			}

			return;
		}		

		// Skip populating the identifier for JPA object copy, as we want this to be null
		// Subclass can override this behavior
		if(isIdentifier(ci)) // handled by processIdentifier method
			return;		

		// If the property to be copied is specified then we perform the copy only for that property
		if(ci.getSourceProperty() != null) { // Owner is not a collection

			// Simply copy the value
			if(ci.isDataType()) {
				ci.setTarget(ci.getTargetObjectCreator().createDataType(ci.getSourceFromParent()));
				Object oldTarget = ci.getTargetFromParent();
				if( !( oldTarget == ci.getTarget() || (oldTarget != null && oldTarget.equals(ci.getTarget()))) ) {
					setModified(ci.getParent().getTarget());
					ci.linkTargetToParent(ci.getTarget());
				}
				return;
			}

			// if a reference was returned then we need to not process it anymore
			ci.setTarget(ci.getTargetObjectCreator().getExistingTarget(ci.getSource())); // TODO: check if processed flag is necessary - What about temp objects with reference relationship?

			if(ci.getTarget() != null) {
				setPropertyTarget(ci, ci.getTarget());

				if( ((Node)ci.getTarget()).isVisited() )
					return;				
			}
			else
				// Do the copy here if necessary
				ci.setTarget(setExistingOrNewCopy(ci));

		}

		// We do not handle collection objects here because they need additional processing on the child elements so, this short circuiting cannot be applied to them.
		if(isExternalAssociationLink(ci)) {
			return;
		}		

		if(ci.getSourceProperty() != null && ci.getSourceProperty().isMany()) // Process the collection/map
			processToMany(ci);		
		else {
			((Node)ci.getTarget()).setVisited(true);
			process(ci);
		}
	}	

	protected boolean isExternalAssociationLink(CallInfo ci) {
		boolean result = ci.isExternal();
		if(ci.getSettings().getDetail() == ContentDetail.TEMPLATE)
			result = !ci.isCascadable(); // TEMPLATE does not include collection references

		return result;
	}	

	protected Object setExistingOrNewCopy(CallInfo ci) throws Exception {
		Object propertyTarget = null;

		Object value = ci.getTargetFromParent();
		if(value != null) {
			// Existing embedded objects need to be wrapped in a data object
			if(ci.getSourceProperty() != null && ((ExtendedType)ci.getSourceProperty().getType()).isEmbedded() && !ExtendedDataObject.class.isAssignableFrom(value.getClass()))
				propertyTarget = createTarget(ci, value, null);
		}

		if(propertyTarget == null)
			// If this is a collection, then we need to skip creating copy if an existing persistent collection is found
			propertyTarget =  createTarget(ci, null); // clear the existing id

		propertyTarget = setPropertyTarget(ci, propertyTarget);

		return propertyTarget;
	}	

	public Object getExistingTarget(CallInfo ci) {
		return ci.getTargetObjectCreator().getExistingTarget(ci.getSource());
	}

	@Override
	public Object createTarget(CallInfo ci, Class<?> desiredClass) {
		return createTarget(ci, null, desiredClass);
	}

	@Override
	public Object createTarget(CallInfo ci, Object targetInstance, Class<?> desiredClass) {
		targetInstance = ClassUtil.getInstance(targetInstance);
		Node target = (Node) ci.getTargetObjectCreator().createTarget(ci, targetInstance, desiredClass);

		return target;
	}	

	protected Object setPropertyTarget(CallInfo ci, Object propertyTarget) throws PropertyMappingException {
		if(ci.getSettings().getAction() == AggregateAction.VERIFY || ci.getStage() != ProcessingStage.UPDATE) // Do not modify if this is just verification
			return propertyTarget;

		// Set the target property value
		ci.linkTargetToParent(propertyTarget);			

		return propertyTarget;
	}		

	protected boolean shouldUnlink(CallInfo ci) {
		return true;	
	}	

	protected boolean isIdentifier(CallInfo ci) {
		if(ci.getSourceProperty() == null)
			return false;

		ExtendedType sourceType = (ExtendedType) ((ExtendedDataObject) ci.getParent().getSource()).getType();
		if(sourceType.isEmbedded()) // embedded type
			return false;

		if(ci.getSourceProperty().getName().equals( sourceType.getIdentifierProperty().getName()) )
			return true;

		return false;
	}

	protected void processNullValue(CallInfo ci) throws Exception {
		if(ci.getSettings().getAction() == AggregateAction.VERIFY || ci.getStage() != ProcessingStage.UPDATE) // Do not modify if this is just verification
			return;

		// Set the null value - unlink the relationship from this side	
		ci.linkTargetToParent(null);				
	}	


	/**
	 * The Data objects for the source collection should have been created
	 * So use the source collection, retrieve the data objects and populate the
	 * target collection
	 * @param object
	 */
	public void cloneToMany(CallInfo callInfo) {
		if(callInfo.getStage() != ProcessingStage.UPDATE)
			return;
		
		Object object = ((ExtendedDataObject) callInfo.getSource()).getInstance();
		java.util.Collection sourceCollection = null;
		java.util.Collection targetCollection = null;
		Map sourceMap = null;
		Map targetMap = null;

		if(callInfo.getSourceProperty().isSet() || callInfo.getSourceProperty().isList()) {
			sourceCollection = (java.util.Collection) object;
			targetCollection = (java.util.Collection) ((ExtendedDataObject) callInfo.getTarget()).getInstance();
			
			for(Object source: sourceCollection) {
				ExtendedDataObject target = (ExtendedDataObject) callInfo.getSourceObjectCreator().getExistingTarget(source);
				targetCollection.add(target.getInstance());
			}			
		} else if(callInfo.getSourceProperty().isMap()) {
			sourceMap = (java.util.Map) object;
			targetMap = (java.util.Map) ((ExtendedDataObject) callInfo.getTarget()).getInstance();
			
			for(Object source: sourceMap.entrySet()) {
				Object targetKey = ((Map.Entry)source).getKey();
				ExtendedDataObject targetValue = (ExtendedDataObject) callInfo.getSourceObjectCreator().getExistingTarget(((Map.Entry)source).getValue());
				targetMap.put(targetKey, targetValue);
			}				
		}
	}

	protected void processCollection(CallInfo callInfo) throws Exception {

		if(callInfo.isDataType())
			return;

		for (Object nextSource : ((ExtendedDataObject)callInfo.getSource()).getList(callInfo)) {
			CallInfo next = new CallInfo(nextSource, null, callInfo, null);
			if(next.isCascadable()) {
				next.setTarget(getExistingTarget(next));
				if( next.getTarget() == null )
					next.setTarget(createTarget(next, null));
				processValue(next);			
			} else
				next.setTarget(createTarget(next, ClassUtil.getInstance(nextSource), null));
		}
		cloneToMany(callInfo);
	}

	@Override
	public boolean preCreatePropertyCallback(CallInfo callInfo) {
		return !callInfo.isValid();
	}	

	protected void customAddToMap(Object targetKey, CallInfo callInfo) {
		if(callInfo.getSettings().getAction() == AggregateAction.VERIFY) // Do not modify if this is just verification
			return;

		Object map                   = callInfo.getParent().getTarget();
		Object targetValue           = callInfo.getTarget();
		ExtendedProperty mapProperty = callInfo.getParent().getTargetProperty();		

		// check if there is a custom collection add
		Method customCollectionAdd = ClassUtil.getCustomAdd(mapProperty.getElementType().getInstanceClass(), mapProperty.getName());
		if(customCollectionAdd != null) {
			CollectionEvent event = new CollectionElement(callInfo, map, targetValue);
			mapProperty.setCustomAdd(map, event);
		} else
			((Map<Object, Object>)map).put(targetKey, targetValue);			
	}		
}
