/**
 * LWSDO, a high performance Domain Driven Design framework.
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * LWSDO is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package org.lwsdo.model.operation;

import javax.persistence.metamodel.Attribute.PersistentAttributeType;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.lwsdo.exception.PropertyMappingException;
import org.lwsdo.model.CallInfo;
import org.lwsdo.model.ExtendedDataObject;
import org.lwsdo.model.ExtendedProperty;
import org.lwsdo.model.ProcessingStage;
import org.lwsdo.model.action.CollectionUpdateAction;
import org.lwsdo.model.action.Executable;
import org.lwsdo.model.action.PropertyKey;
import org.lwsdo.model.action.RemoveElementAction;
import org.lwsdo.model.action.SetterAction;


public class ModifyOperation extends AbstractOperation {
	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());
	
	private Object result;

	public void removeObsoleteEntities(CallInfo callInfo, CollectionUpdateAction migratorAction) {

		// If we are merging then obsolete elements are not removed 		
		if(callInfo.getSettings().doMerge(callInfo))
			return;		

		migratorAction.removeObsoletes(callInfo);

	}	

	@Override
	public void processCollection(CallInfo callInfo) throws Exception {

		if(!callInfo.isCascadable()) // Hibernate will not update this collection
			return;
		
		if(callInfo.getStage() != ProcessingStage.UPDATE)
			return;		

		/* Check and see if there is a collection key, if not use the id to match the original and copy 
		 * Get the method that gives this information
		 * populate the jPAObjects map using the key (id or userKey)
		 * Iterate through the original collection 
		 *   if it is an existing object - update it
		 *   if it is a new object - create and update it
		 */
		PropertyKey propertyKey = new PropertyKey((ExtendedDataObject) callInfo.getParent().getTarget(), callInfo.getTargetProperty());
		CollectionUpdateAction migratorAction = callInfo.getTargetRoot().getDependencyGraph().getOrCreateMigratorAction(propertyKey);

		removeObsoleteEntities(callInfo, migratorAction);
		migratorAction.addOrReposition(callInfo);
	}			

	@Override
	protected boolean shouldUnlink(CallInfo ci) {

		if(ci.getSource() == null || ci.getSourceProperty() == null)
			return true;

		return ci.getSourceProperty().isTransient();
	}	

	@Override
	protected void processNullValue(CallInfo ci) throws Exception {

		if(ci.getSettings().doUpdate(ci) && ci.getStage() == ProcessingStage.UPDATE) {
			
			if(ci.getSourceProperty() != null && ci.getSourceProperty().isMany()) {
				ci.setTarget(createTarget(ci, ci.getTargetFromParent(), null)); // Get the persistent data object into target
				processCollection(ci);
			} else {
				unlinkOneToOne(ci);
				unlinkManyToOne(ci);
			}
		}
	}	
	
	protected void unlinkManyToOne(CallInfo ci) {
		if(ci.getTargetProperty().getAssociationType() != PersistentAttributeType.MANY_TO_ONE)
			return;

		// If this is a uni-directional OneToOne, we just have to set the value to null
		if(!ci.getTargetProperty().isBiDirectional()) {
			// Set the null value	
			ci.linkTargetToParent(null);				
		} else { // bi-directional link
			// If the existing value is null then return
			if(ci.getTargetFromParent() == null)
				return;

			ci.setTarget(createTarget(ci, ci.getTargetFromParent(), null)); // Get the persistent data object into target

			PropertyKey key = new PropertyKey((ExtendedDataObject)ci.getParentTargetEntity(), ci.getTargetProperty());
			PropertyKey oppositeKey = new PropertyKey((ExtendedDataObject)ci.getTarget(), ci.getTargetProperty().getOpposite());

			// Set the target to null
			Executable originalAction = new SetterAction(null, key, null);
			ci.getTargetRoot().getDependencyGraph().addAction(originalAction);

			// set the backRef from the opposite object to null	
			Object position = (oppositeKey.getProperty() == null) ? null : ((ExtendedProperty)oppositeKey.getProperty()).getValue(key.getDataObject());
			RemoveElementAction removeAction = new RemoveElementAction(oppositeKey, key.getDataObject(), ci, originalAction, position);
		}
	}	

	protected void unlinkOneToOne(CallInfo ci) {
		if(ci.getTargetProperty().getAssociationType() != PersistentAttributeType.ONE_TO_ONE)
			return;

		// If this is a uni-directional OneToOne, we just have to set the value to null
		if(!ci.getTargetProperty().isBiDirectional()) {
			// Set the null value	
			ci.linkTargetToParent(null);				
		} else { // bi-directional link
			// If the existing value is null then return
			if(ci.getTargetFromParent() == null)
				return;

			ci.setTarget(createTarget(ci, ci.getTargetFromParent(), null)); // Get the persistent data object into target

			PropertyKey key = new PropertyKey((ExtendedDataObject)ci.getParentTargetEntity(), ci.getTargetProperty());
			PropertyKey oppositeKey = new PropertyKey((ExtendedDataObject)ci.getTarget(), ci.getTargetProperty().getOpposite());

			// Set the target to null
			Executable originalAction = new SetterAction(null, key, null);
			ci.getTargetRoot().getDependencyGraph().addAction(originalAction);

			// set the backRef from the opposite object to null			
			ci.getTargetRoot().getDependencyGraph().addAction(new SetterAction(null, oppositeKey, originalAction));
		}
	}

	protected void linkOneToOne(CallInfo ci, Object value) {
		if(ci.getTargetProperty().getAssociationType() != PersistentAttributeType.ONE_TO_ONE)
			return;

		// If this is a uni-directional OneToOne, we just have to set the target to value
		if(!ci.getTargetProperty().isBiDirectional()) {
			// Set the value	
			ci.linkTargetToParent(value);				
		} else {
			PropertyKey key            = new PropertyKey((ExtendedDataObject)ci.getParentTargetEntity(), ci.getTargetProperty());
			PropertyKey newOppositeKey = new PropertyKey((ExtendedDataObject)value, ci.getTargetProperty().getOpposite());	
			
			Executable originalAction = new SetterAction(value, key, null);
			// Set the target to value
			ci.getTargetRoot().getDependencyGraph().addAction(originalAction);			

			// If the target has an existing opposite object, its backRef needs to be set to null 
			if(ci.getTargetFromParent() != null) {
				CallInfo newCall = new CallInfo(null, null, ci, ci.getTargetProperty());
				Object oldObject = createTarget(newCall, ci.getTargetFromParent(), null); // Get the persistent data object into target

				// Set the backRef of the opposite object to null
				PropertyKey oldOppositeKey = new PropertyKey((ExtendedDataObject)oldObject, ci.getTargetProperty().getOpposite());

				// Set the backRef from the old value to null
				ci.getTargetRoot().getDependencyGraph().addAction(new SetterAction(null, oldOppositeKey, originalAction));
			}			 

			// Set the backRef of value to point to the parent target (or to the ancester in case of embedded property)
			ci.getTargetRoot().getDependencyGraph().addAction(new SetterAction(ci.getParentTargetEntity(), newOppositeKey, originalAction)); 				

		}
	}

	protected Object setPropertyTarget(CallInfo ci, Object propertyTarget) throws PropertyMappingException {
		if(ci.getStage() != ProcessingStage.UPDATE)
			return propertyTarget;

		// TODO: If the property is a mapKey or an indexKey then the toMany side needs to know about the reposition action 
		
		if(ci.getTargetProperty().isMany())
			super.setPropertyTarget(ci, propertyTarget);
		else
			linkOneToOne(ci, propertyTarget);	
		// TODO: MANY_TO_ONE

		return propertyTarget;
	}

	@Override
	protected void setResult(Object target) {
		this.result = target;
	}	

	@Override
	public Object getResult() {
		return result;
	}	
}
