/**
 * LWSDO, a high performance Domain Driven Design framework.
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * LWSDO is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package org.lwsdo.model.view;

import java.util.Map;
import java.util.TreeMap;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.lwsdo.model.ExtendedDataObject;
import org.lwsdo.model.Settings;
import org.lwsdo.service.QueryCapability;

/**
 * TODO: QueryBuilder needs to handle the following expressions

First Phase -- Restrict to what is needed by CSA 

CSA has the following filters
-----------------------------
ilike(param, %1)
in(param, %1)
idEq(%1)
ge(%1)
asc(%1)

e.g.

ilike(name, "%xyz%")
in("id", String[])
idEq(String)
ge("createdOn", long)
ge("updatedOn", long)
asc("name")

Modify view to support filters
------------------------------

ARTIFACTFILTER
id
name
displayName
description
iconUrl
detailedDescription
isCriticalSystemObject
ownedBy.name
state.name
artifactType.name
disabled
  FILTERS
  ilike(name, ":name")            // lower(name) LIKE :name
  in(state, ":state"              // name IN(:state)
  equal(ownedBy.name, ":owner")   // ownedBy.name = :owner 
  ge(createdOn, ":createdSince")  // createdOn > :createdSince
  ge(updatedOn, ":updatedSince"); // updaetdOn > :updatedSince
  asc(name)                       // ORDER BY name  also honor OrderBy annotation in model if no ORDER BY clause is provided

 *
 */
public abstract class QueryBuilder {
	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());

	private static final String SELECT_CLAUSE = "SELECT ";

	private ExtendedDataObject entity;
	private ViewBranch view;
	private boolean SQL;

	public abstract Query createQuery(String hql);

	public boolean isSQL() {
		return SQL;
	}

	public void init(ExtendedDataObject entity, ViewBranch view) {
		this.entity = entity;
		this.view = view;
	}

	private QueryCapability getQueryCapability()  {
		return entity.getObjectCreator().getDAS().getPersistenceOrchestrator().getQueryCapability();
	}

	public String constructQuery(Map<String, String> filters) {
		//		First check for a SQL query 
		//		When retrieving the QueryBuilder instance, registered SQL queries are given preference over HQL/JPQL queries.
		//
		//		It is the responsibility of the user to ensure that the registered SQL is ANSI compliant 
		//		and can work with all the databases that the user uses.

		// TODO: handle sub-branches

		if(view.getContentView().getNativeQuery() != null) {
			SQL = true;
			String queryString = view.getContentView().getNativeQuery().getQueryString();
			return queryString.replaceAll("[\n\r]", "");
		}

		StringBuffer HQL = new StringBuffer(constructHQL(filters));
		HQL.append(buildWhereClause(filters));
		HQL.append(buildOrderClause());

		return HQL.toString();		
	}

	protected String constructHQL(Map<String, String> filters) {
		Map<String, ColumnMeta> meta = view.getAugmentedAttributes();

		StringBuffer HQL = new StringBuffer(SELECT_CLAUSE);

		int position = 0;
		for(ColumnMeta columnMeta: meta.values()) {
			columnMeta.setPosition(position++);
			if(!SELECT_CLAUSE.equals(HQL.toString()))
				HQL.append(", ");
			String columnString = columnMeta.getQueryString(getQueryCapability());
			HQL.append(columnString);
		}

		HQL.append(" FROM " + entity.getType().getInstanceClass().getName() + " AS " + view.getViewProperty(ViewProperty.ROOT_PROPERTY_NAME).getAlias());

		// TODO: if more than one collection property is specified then we would get a severe performance hit due to cartesian product
		//       This can be solved by breaking the query intelligently. This is for the future, for now we shall log a warning
		// getCollectionBranches()

		for(ViewProperty viewProperty: view.getAliasedItems()) {
			ViewProperty parentView = viewProperty.getParent();
			if(parentView == null)
				continue; // the root needs to be skipped
			HQL.append(" LEFT OUTER JOIN " + viewProperty.getNormalizedName() + " AS " + viewProperty.getAlias());		
		}

		return HQL.toString();
	}	

	protected String buildOrderClause() {
		StringBuffer result = new StringBuffer();
		
		Map<Integer, String> orderClauses = new TreeMap<Integer, String>();
		for(ColumnMeta columnMeta: view.getAugmentedAttributes().values()) {
			if(columnMeta.getAttributePath().endsWith(ViewProperty.LIST_INDEX_ATTRIBUTE)) {
				String columnString = columnMeta.getQueryString(getQueryCapability());
				orderClauses.put(columnMeta.getAttributeLevel(), columnString);
			}
		}

		// Append any ORDER BY clauses
		if(orderClauses.size() > 0) {
			result.append(" ORDER BY ");
			StringBuffer orderString = new StringBuffer();
			for(String indexColumn: orderClauses.values())
				orderString.append( (orderString.length() > 0) ? ", " + indexColumn : indexColumn );
			result.append(orderString);
		}		
		
		return result.toString();
	}

	protected String getAliasedName(String propertyPath) {
		ViewProperty viewProperty = view.getViewProperty(propertyPath);
		return viewProperty.getNormalizedName();
	}	

	protected String buildWhereClause(Map<String, String> filters) {
		StringBuffer result = new StringBuffer(" WHERE 1 = 1 ");

		for(String key: filters.keySet())
			result.append(" AND " + getAliasedName(key) + " = :" +  Settings.encodeParam(key));

		return result.toString();
	}	
}
