/**
 * LWSDO, a high performance Domain Driven Design framework.
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * LWSDO is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package org.lwsdo.model.view;

import java.util.ArrayList;
import java.util.List;

import org.lwsdo.model.ExtendedProperty;
import org.lwsdo.model.ExtendedType;
import org.lwsdo.model.Property;
import org.lwsdo.model.Settings;
import org.lwsdo.model.Type;

public class ViewProperty {
	public static final String ROOT_PROPERTY_NAME = "root";	
	
	public static final String ENTITYNAME_ATTRIBUTE  = "TYPE_";
	public static final String MAP_KEY_ATTRIBUTE     = "KEY_";
	public static final String MAP_VALUE_ATTRIBUTE   = "VALUE_";
	public static final String LIST_INDEX_ATTRIBUTE  = "INDEX_";               // Use this to set the object at the right location or in ORDER BY clause if filter is used	
	public static final String COL_USERKEY_ATTRIBUTE = "COLLECTION_USERKEY_";
	public static final String USERKEY_ATTRIBUTE     = "USERKEY_";

	private ViewProperty   parent;        // Needed for embedded or element collection attributes
	private Property       property;      // null for root
	private Type           type;          // the type of the entity
	private String         propertyPath;  // the property path that represents the attribute
	private boolean        isDynamic;     // Was not part of the initial set of properties when this view was created
	private String         alias;         // Refers to the entity/collection entity alias
	private String         propertyAlias; // NOTE: A property cannot typically be used in a where clause	

	public String getPropertyAlias() {
		return propertyAlias;
	}

	public void setPropertyAlias(String propertyAlias) {
		this.propertyAlias = propertyAlias;
	}

	public ViewProperty getParent() {
		return parent;
	}

	public boolean isDynamic() {
		return isDynamic;
	}

	public Property getIdentifierProperty() {
		if(type.isDataType())
			throw new IllegalArgumentException("This method can only be invoked on a data object");

		return ((ExtendedType)type).getIdentifierProperty();
	}

	public String getPropertyPath() {
		return propertyPath;
	}

	public boolean isCollection() {
		return (property != null) ? property.isMany() : false;
	}

	private void setProperty() {
		if(parent == null) { 
			if(propertyPath.equals(ROOT_PROPERTY_NAME))
				return;
			else
				throw new IllegalStateException("anchor is not set for non-root property");
		}

		property = parent.getProperty(propertyPath);
		if(property != null) {
			this.type = this.property.getType();
			if(property.isMany()) {
				this.type = ((ExtendedProperty)property).getElementType();
			}				
		}
	}

	public Type getType() {
		return this.type;
	}

	public ViewProperty(boolean isDynamic, Type type) {
		this.propertyPath = ROOT_PROPERTY_NAME;
		this.isDynamic = isDynamic;
		this.type = type;
		this.parent = null;
	}

	public ViewProperty(String attribute, boolean isDynamic, ViewProperty parent) {
		this.propertyPath = qualifyProperty(attribute);
		this.isDynamic = isDynamic;
		this.parent = parent;

		setProperty();
	}
	
	public static String qualifyProperty(String propertyPath) {
		if(!propertyPath.startsWith(ROOT_PROPERTY_NAME))		
			return ROOT_PROPERTY_NAME + Settings.PATH_DELIMITER + propertyPath;
		else
			return propertyPath;
	}

	public static String unqualifyProperty(String propertyPath) {
		return propertyPath.substring(ROOT_PROPERTY_NAME.length() + Settings.PATH_DELIMITER.length());
	}	

	public static boolean isQualified(String propertyPath) {
		if(!propertyPath.startsWith(ROOT_PROPERTY_NAME))		
			return false;
		else
			return true;
	}	
	
	public static String getBaseName(String propertyPath) {
		if(propertyPath.indexOf(Settings.PATH_DELIMITER) != -1)
			return propertyPath.substring(propertyPath.lastIndexOf(Settings.PATH_DELIMITER)+1);
		else
			return propertyPath;
	}	
	
	public static String getRootName(String propertyPath) {
		if(propertyPath.indexOf(Settings.PATH_DELIMITER) != -1)
			return propertyPath.substring(0, propertyPath.indexOf(Settings.PATH_DELIMITER));
		else
			return propertyPath;
	}	
	
	public static String getNext(String propertyPath) {
		if(propertyPath.indexOf(Settings.PATH_DELIMITER) != -1)
			return propertyPath.substring(propertyPath.indexOf(Settings.PATH_DELIMITER)+1);
		else
			return null;
	}	
	
	public String getAlias() {
		return alias;
	}

	public boolean isEntity() {
		if(type != null && ExtendedType.class.isAssignableFrom(type.getClass()) && ((ExtendedType)type).getIdentifierProperty() != null )
			return true;

		return false;
	}

	public void setAlias(String alias) {
		if(!isEntity())
			throw new IllegalArgumentException("Alias can only be set for an entity or a collection of entities");

		this.alias = alias;
	}

	public Property getProperty() {
		return property;
	}	
	
	/**
	 * Normalize the property name based on any aliases in the ancestor list.
	 * This is needed for HQL to work properly 
	 * @return
	 */
	protected String getNormalizedName() {

		String baseName = getBaseName(propertyPath);
	
		if(getParent() == null)
			return baseName;
		else
			return getParent().getAlias() + Settings.PATH_DELIMITER + baseName;
	}	

	/*
	 * Retrieves the property specified in propertyPath
	 */
	public Property getProperty(String childPath) {
		String path = isQualified(childPath) ? null : childPath;
		if(childPath.startsWith(this.propertyPath))
			path = childPath.substring(this.propertyPath.length() + Settings.PATH_DELIMITER.length());

		if(path == null)
			return null;

		return this.type.getProperty(path);
	}		

	public void setProperty(Property property) {
		this.property = property;
	}

	public List<ColumnMeta> getColumnMeta(boolean narrow) {
		List<ColumnMeta> result = new ArrayList<ColumnMeta>();

		if(!isEntity()) {
			result.add(new ColumnMeta(propertyPath, this));
		} else {
			// Add the entity name
			if(narrow)
				result.add(new ColumnMeta(propertyPath + Settings.PATH_DELIMITER + ViewProperty.ENTITYNAME_ATTRIBUTE, this));
			
			// Try to add the identifier always if possible. This will be replaced by the viewProperty which is not dynamic in the QueryBuilder
			if( ((ExtendedType)type).getIdentifierProperty() != null )
				result.add(new ColumnMeta(propertyPath + Settings.PATH_DELIMITER + ((ExtendedType)type).getIdentifierProperty().getName(), this));			
			if(property != null && property.isMany()) { // collection
				if( ((ExtendedProperty)property).isList() ) {
					// add INDEX
					result.add(new ColumnMeta(propertyPath + Settings.PATH_DELIMITER + LIST_INDEX_ATTRIBUTE, this));
				} else if( ((ExtendedProperty)property).isMap() ) {
					// add KEY
					result.add(new ColumnMeta(propertyPath + Settings.PATH_DELIMITER + MAP_KEY_ATTRIBUTE, this));
					// add COLLECTION_USERKEY
					if( ((ExtendedType)type).getCollectionUserKey() != null )
						result.add(new ColumnMeta(propertyPath + ((ExtendedType)type).getCollectionUserKey().getName(), this));
				}
			} else { // entity
				// add USERKEY
				if( ((ExtendedType)type).getUserKey() != null )
					result.add(new ColumnMeta(propertyPath + ((ExtendedType)type).getUserKey().getName(), this));
			}
		}

		return result;
	}

}
