/**
 * LWSDO, a high performance Domain Driven Design framework.
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * LWSDO is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package org.lwsdo.util;

import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.Modifier;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

import org.apache.commons.lang.ClassUtils;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.lwsdo.model.AggregateAction;
import org.lwsdo.model.CallInfo;
import org.lwsdo.model.EntityKey;
import org.lwsdo.model.ExtendedDataObject;
import org.lwsdo.model.ExtendedDataObjectImpl;
import org.lwsdo.model.ExtendedProperty;
import org.lwsdo.model.ExtendedType;
import org.lwsdo.model.MapperDirection;
import org.lwsdo.model.Node;
import org.lwsdo.model.Property;
import org.lwsdo.model.SimpleType;
import org.lwsdo.model.Type;
import org.lwsdo.model.TypeMapper;
import org.lwsdo.service.DataAccessService;

/**
 * This class is used to create the appropriate instances of the copy, given the
 * class of the original object.
 * 
 * @author Dilip Dalton
 */
public class ObjectCreator {
	private static final Logger logger = LogManager.getLogger(new Exception()
	.getStackTrace()[0].getClassName());

	// This map is used to prevent multiple copies of the same object.
	private Map<Object, Object>                sourceTargetMap = new IdentityHashMap<Object, Object>();
	private Map<EntityKey, ExtendedDataObject> entitiesByKey = new HashMap<EntityKey, ExtendedDataObject>();
	private DataAccessService                  das;
	private TypeMapper                         typeMapper;

	public ObjectCreator(DataAccessService das, MapperDirection direction) {
		this.das = das;
		this.typeMapper = das.getTypeMapper().newInstance(direction);
	}

	public DataAccessService getDAS() {
		return das;
	} 

	public TypeMapper getTypeMapper() {
		return typeMapper;
	}
	
	public void addByEntityKey(EntityKey key, ExtendedDataObject entity) {
		entitiesByKey.put(key, entity);
	}
	
	public void removeByEntityKey(EntityKey key) {
		entitiesByKey.remove(key);
	}	
	
	public ExtendedDataObject getByEntityKey(EntityKey key) {
		return entitiesByKey.get(key);		
	}	

	public List<ExtendedDataObject> getDataObjects() {
		List<ExtendedDataObject> result = new ArrayList<ExtendedDataObject>();

		for(Object potentialDataObject: sourceTargetMap.values()) {
			if(ExtendedDataObject.class.isAssignableFrom(potentialDataObject.getClass())) 
				result.add((ExtendedDataObject) potentialDataObject);
		}

		return result;
	}

	public void clearState() {
		sourceTargetMap = new IdentityHashMap<Object, Object>(); // Map of the source instance and the target DataObject
	}

	/**
	 * Returns true if the given class is under a package that starts with
	 * "java.".
	 */
	protected boolean isJavaOrAppPackage(Class<?> c) {
		if (c == null)
			return false;
		Package p = c.getPackage();
		return p != null
				&& (p.getName().startsWith("java.") || p.getName().startsWith(
						"org.LWSDO."));
	}

	/**
	 * Returns true if the given class is known to be immutable; false
	 * otherwise.
	 */
	protected boolean immutable(Class<?> c) {
		if (c == null)
			return false;
		return c == String.class || c.isPrimitive() || c.isEnum()
				|| Number.class.isAssignableFrom(c)
				|| Boolean.class == c;
	}

	public Object getExistingTarget(Object source) {
		if(source == null)
			return null;

		// The source is based on the instance object so extract if necessary 
		if(ExtendedDataObject.class.isAssignableFrom(source.getClass()))
			source = ((ExtendedDataObject)source).getInstance();

		return sourceTargetMap.get(source);
	}

	public void updateExistingTarget(Object source, Object target) {
		sourceTargetMap.put(source, target);
	}	

	public final void reset() {
		sourceTargetMap = new IdentityHashMap<Object, Object>();		
	}

	/*
	 * for now let us assume that a class with only private constructors is
	 * immutable. Later on we shall enhance this using annotations e.g., the
	 * immutable property in hibernate mapping
	 */
	private boolean hasOnlyPrivateConstructors(Class<?> sourceClass) {
		boolean result = true;

		for (Constructor<?> con : sourceClass.getDeclaredConstructors()) {
			final int mod = con.getModifiers();
			if (!Modifier.isPrivate(mod)) {
				result = false;
				break;
			}
		}

		return result;
	}

	public Object createDataType(Object instance) throws Exception {
		return newInstance(instance, instance.getClass());		
	}

	public ExtendedDataObject createDataObject(Object targetInstance, Type targetType, ExtendedDataObject container, Property containmentProperty) {
		ExtendedDataObject result = null;

		if(ExtendedDataObject.class.isAssignableFrom(targetInstance.getClass()))
			throw new RuntimeException("Instance is a Data Object");

		if(targetInstance != null)
			result = (ExtendedDataObject) getExistingTarget(targetInstance);

		if(result == null)
			result = this.createDataObject(null, targetInstance, targetType, container, containmentProperty);

		return result;
	}

	/**
	 * This method ensures that all derived and reference persistence instance point to the same reference data object 
	 * @param sourceInstance
	 * @param targetInstance
	 * @param targetType
	 * @param container
	 * @param containmentProperty
	 * @return
	 */
	private ExtendedDataObject createDataObject(Object sourceInstance, Object targetInstance, Type targetType, ExtendedDataObject container, Property containmentProperty) {

		ExtendedDataObject dataObject = (ExtendedDataObject) sourceTargetMap.get(targetInstance);
		if(dataObject != null && sourceInstance != null) {
			sourceTargetMap.put(sourceInstance, dataObject);
			return dataObject;
		}

		dataObject = new ExtendedDataObjectImpl(targetType,
				container,
				containmentProperty,
				this); 
		dataObject.setInstance(targetInstance);	

		dataObject = register(dataObject);

		if(sourceInstance != null)
			sourceTargetMap.put(sourceInstance, dataObject);

		return dataObject;
	}

	public void unregister(ExtendedDataObject dataObject) {
		sourceTargetMap.remove(dataObject.getInstance());
		dataObject.removeEntity(dataObject);
	}

	public ExtendedDataObject register(ExtendedDataObject newDataObject) {
		ExtendedDataObject result = newDataObject;

		Object instance = newDataObject.getInstance();
		if(instance == null) 
			throw new IllegalStateException("No persistent state found - possible data corruption");

		sourceTargetMap.put(instance, newDataObject);

		// If an existing persistent object is already present, check if it is of a more general type, if so it has to be replaced
		// with the current object
		ExtendedDataObject existing = newDataObject.getEntity(newDataObject);
		if(existing != newDataObject) {
			if(existing != null) {
				ExtendedType existingRootType = ((ExtendedType)existing.getType()).getRootEntityType();
				ExtendedType newRootType = ((ExtendedType)newDataObject.getType()).getRootEntityType();
				if(existingRootType == newRootType) {
					throw new IllegalStateException("Creating two data objects for the same id and root type");
				} else if(existingRootType.getInstanceClass().isAssignableFrom(newRootType.getInstanceClass())) {
					// New data object is of a more specific type (This could be a proxy created by the persistent mechanism)
					Object oldInstance = existing.getInstance();
					existing.setInstance(instance);

					// Clear up references to the more general proxy instance
					sourceTargetMap.remove(oldInstance);
					((Node)existing).replaceInstance(instance);
					result = existing; // we have replace the instance with the new instance
				} else {
					result = existing; // new data object is of a more general type, so we use the existing data object				
				}
			} else {
				// Register if it is a persistent instance
				newDataObject.addEntity(newDataObject);
			}
		} else {
			throw new IllegalStateException("New data object was registered by a different means!");
		}

		// Do the actual registration
		sourceTargetMap.put(result.getInstance(), result);

		return result;
	}

	/**
	 * Create a new instance based on the type of conversion needed on the "from" object
	 * @param source
	 * @param typeMapper
	 * @return
	 * @throws Exception
	 */
	public Object createTarget(CallInfo ci, Object targetInstance, Class<?> desiredClass) {
		Object source = ClassUtil.getInstance(ci.getSource());
		ExtendedDataObject targetContainer = (ci.getParent() == null) ? null : (ExtendedDataObject)ci.getParent().getTarget();
		Property sourceContainmentProperty = ci.getSourceProperty();

		Object result = getExistingTarget(source);
		if(result != null && (targetInstance != null && targetInstance != ClassUtil.getInstance(result)) )
			throw new IllegalStateException("existing target instance not the same as the passed in target instance");

		if(targetInstance != null) {
			Object potentialResult = getExistingTarget(targetInstance);
			if(potentialResult != null && result != null && potentialResult != result)
				throw new IllegalStateException("Source and target instance data objects are different");

			if(result == null && potentialResult != null) {
				if(source != null)
					sourceTargetMap.put(source, potentialResult);
				result = potentialResult;
			}
		}

		try {
			Class<?> targetClass = null;
			if(targetInstance == null) {
				targetClass = source.getClass();

				if(desiredClass == null) {
					if(ExtendedDataObject.class.isAssignableFrom(source.getClass())) {
						ExtendedType type = (ExtendedType) ((ExtendedDataObject)source).getType();
						targetClass = typeMapper.getClass(type.getInstanceClass());
					}
				} else
					targetClass = desiredClass;
			} else
				targetClass = targetInstance.getClass();

			Type targetType = das.getType(targetClass);
			Property targetContainmentProperty = (targetContainer == null || sourceContainmentProperty == null) ? null : targetContainer.getInstanceProperty(sourceContainmentProperty.getName()); 	

			/* 
			 * We need to fetch a persistent instance for the following:
			 * 1. AggregateAction is CREATE or CLONE and the object being fetched is not part of the aggregate
			 * 2. Any UPDATE or MERGE action
			 */
			AggregateAction action = ci.getSettings().getAction();
			boolean getPersistent =  ( (action == AggregateAction.CREATE || action == AggregateAction.CLONE) && !ci.isCascadable() ) ||
					(action == AggregateAction.MERGE || action == AggregateAction.UPDATE);
			if( getPersistent && targetInstance == null) {

				if(result == null) {
					targetInstance = das.getPersistentObject(ci, typeMapper);
					if(!ci.isDataType() && ci.getSourceObjectCreator().getExistingTarget(targetInstance) != null)
						throw new IllegalArgumentException("Cannot use a persistence managed object as input");
					if(targetInstance != null) {
						result = createDataObject(source, targetInstance, targetType, targetContainer, targetContainmentProperty);
						((ExtendedDataObject)result).setPersistent(true);
						((ExtendedDataObject)result).addNode();
					}
				}
			} 

			if(result == null) {
				if(targetInstance == null)
					targetInstance = createInstance(source, targetClass, targetType);
				result = createDataObject(source, targetInstance, targetType, targetContainer, targetContainmentProperty);
				if( !(ci.isDataType() || ci.getSettings().getAction() == AggregateAction.READ) ) // both conditions should be false
					((ExtendedDataObject)result).addNode();
			}

			// Give opportunity for reference objects to have any post processing done by the DAS
			das.postProcess(targetInstance);
		} catch (Exception e) {
			throw ClassUtil.wrapRun(e);
		}	

		return result;
	}	

	protected void setEmbeddedInstance(Object instance, Type targetType) throws Exception {
		for(Property property: targetType.getProperties()) {
			if(SimpleType.class.isAssignableFrom(property.getType().getClass()))
				continue;
			ExtendedType propertyType = (ExtendedType) property.getType();
			if(propertyType.isEmbedded()) {
				Object embeddedInstance = newInstance(propertyType.getInstanceClass());
				((ExtendedProperty)property).setValue(instance, embeddedInstance);
				setEmbeddedInstance(embeddedInstance, propertyType); // recurse if necessary
			}
		}		
	}

	protected Object createInstance(Object source, Class<?> targetClass, Type targetType) throws Exception {
		Object instance = newInstance(source, targetClass);
		setEmbeddedInstance(instance, targetType);

		return instance;
	}

	public Object createInstance(Type targetType) throws Exception {
		Object instance = newInstance(null, targetType.getInstanceClass());
		setEmbeddedInstance(instance, targetType);

		return instance;
	}	

	public Object createTarget(CallInfo ci) {
		return createTarget(ci, null, null);
	}	

	public Object createTarget(CallInfo ci, Class<?> desiredClass) {
		return createTarget(ci, null, desiredClass);
	}

	/**
	 * Create an object based on the class off the "from" object.
	 * 
	 * @param from
	 *            A new new instance based on this object's class is created
	 * @param toClass
	 *            This is used for validation, to ensure that the new instance
	 *            can be assigned to this type
	 * @return new instance
	 */
	protected Object newInstance(Object from, Class<?> toClass) throws Exception {

		// Not necessary to create a new instance for an immutable object
		if (immutable(toClass) || (from != null && hasOnlyPrivateConstructors(from.getClass())) )
			return from;

		Class<?> fromClass = from != null ? from.getClass() : toClass;

		// Collection
		if (Collection.class.isAssignableFrom(fromClass)) {
			if (!toClass.isAssignableFrom(fromClass))
				return null;
			return newCollection((Collection<?>) from, fromClass);
		}

		// Map
		if (Map.class.isAssignableFrom(fromClass)) {
			if (!toClass.isAssignableFrom(fromClass))
				return null;
			return newMap((Map<?, ?>) from, fromClass);
		}

		// Array
		if (fromClass.isArray()) {
			if (!toClass.isAssignableFrom(fromClass))
				return null;
			return newArray(from);
		}

		// Make copies of the following object
		// Timestamp
		if (Timestamp.class.isAssignableFrom(fromClass)) {
			if (!toClass.isAssignableFrom(fromClass))
				return null;
			Timestamp ts = (Timestamp) from;
			return new Timestamp(ts.getTime());
		}
		// Date
		if (Date.class.isAssignableFrom(fromClass)) {
			if (!toClass.isAssignableFrom(fromClass))
				return null;
			Date date = (Date) from;
			return new Date(date.getTime());
		}

		String fromPackageName = ClassUtils.getPackageName(fromClass);
		if (fromPackageName.startsWith("java.")) {
			if (!toClass.isAssignableFrom(fromClass))
				return null;
		}

		// We cannot infer the appropriate "from" target class, so we fallback to "toClass"
		return newApplicationObject(from, toClass);
	}

	private Object newApplicationObject(Object from, Class<?> toClass) throws Exception {
		Object to = newInstance(toClass);

		return to;
	}

	private Object newArray(Object from) {
		Class<?> fromClass = from.getClass();
		Class<?> fromComponentType = fromClass.getComponentType();

		// primitive array
		if (immutable(fromComponentType)) {
			int len = Array.getLength(from);
			Object toArray = Array.newInstance(fromComponentType, len);
			return toArray;
		}

		// non-primitive array
		int len = Array.getLength(from);
		Object toArray = Array.newInstance(fromComponentType, len);

		return toArray;
	}

	@SuppressWarnings("unchecked")
	private Collection<?> newCollection(Collection<?> from, Class<?> fromClass) throws Exception {
		Collection<Object> toCollection = (Collection<Object>) ClassUtil.getInstance(sourceTargetMap.get(from));

		if (toCollection == null)
			toCollection = (Collection<Object>) this.createCollection(from, fromClass);

		return toCollection;
	}

	@SuppressWarnings("unchecked")
	private Map<?, ?> newMap(Map<?, ?> from, Class<?> fromClass) throws Exception {
		Map<Object, Object> toMap = (Map<Object, Object>) ClassUtil.getInstance(sourceTargetMap.get(from));

		if (toMap == null)
			toMap = this.createMap(from, fromClass);

		return toMap;
	}

	/**
	 * Returns a replicated comparator of the given sorted map, or null if there
	 * is no comparator.
	 */
	@SuppressWarnings("unchecked")
	private Comparator<? super Object> newComparator(
			SortedMap<?, ?> fromSortedMap) throws Exception {
		if(fromSortedMap == null)
			return null;
		
		Comparator<?> fromComparator = fromSortedMap.comparator();
		Comparator<?> toComparator = fromComparator == null ? null
				: (Comparator<?>) this.newApplicationObject(fromComparator,
						Comparator.class);
		return (Comparator<? super Object>) toComparator;
	}

	/**
	 * Returns a replicated comparator of the given sorted set, or null if there
	 * is no comparator.
	 */
	private Comparator<?> newComparator(SortedSet<?> fromSortedSet) throws Exception {
		if(fromSortedSet == null)
			return null;
		
		Comparator<?> fromComparator = fromSortedSet.comparator();
		Comparator<?> toComparator = fromComparator == null ? null
				: (Comparator<?>) this.newApplicationObject(fromComparator,
						Comparator.class);
		return toComparator;
	}

	private boolean useJavaType(Class<?> fromClass) {
		boolean canInstantiate = false;
		for (Constructor<?> con : fromClass.getDeclaredConstructors()) {
			if (con.getParameterTypes().length == 0) {
				canInstantiate = true;
				break;
			}
		}

		return !isJavaOrAppPackage(fromClass) || !canInstantiate;
	}

	/**
	 * 
	 * @param from If this argument is null then a new object based on fromClass will be created
	 * @param fromClass
	 * @return
	 * @throws Exception
	 */
	// We also need to create a new instance for any user created comparator objects
	private Object createCollection(Collection<?> from, Class<?> fromClass) throws Exception {

		if (SortedSet.class.isAssignableFrom(fromClass)) {
			SortedSet<?> fromSortedSet = (SortedSet<?>) from;
			Comparator toComparator = newComparator(fromSortedSet);

			if (!isJavaOrAppPackage(fromClass))
				return new TreeSet<Object>(toComparator);
			Constructor<?> constructor = fromClass.getConstructor(Comparator.class);
			Object[] initargs = { toComparator };
			return constructor.newInstance(initargs);
		}
		if (Set.class.isAssignableFrom(fromClass)) {
			if (useJavaType(fromClass))
				return new HashSet<Object>();
			return fromClass.newInstance();
		}
		if (List.class.isAssignableFrom(fromClass)) {
			if (useJavaType(fromClass))
				return new ArrayList<Object>();
			return fromClass.newInstance();
		}
		if (isJavaOrAppPackage(fromClass) && !fromClass.isInterface()) {
			if (SortedSet.class.isAssignableFrom(fromClass)) {
				SortedSet<?> fromSortedSet = (SortedSet<?>) from;
				Comparator<?> toComparator = from != null ? newComparator(fromSortedSet) : null;

				if (toComparator != null)
					return newInstanceWithComparator(fromSortedSet.getClass(),
							toComparator);
			}
			return newInstance(fromClass);
		}

		// Default collection
		return new ArrayList<Object>();
	}

	/**
	 * 
	 * @param from If this argument is null, then a new object based on fromClass will be created
	 * @param fromClass
	 * @return
	 * @throws Exception
	 */
	private Map<Object, Object> createMap(Map<?, ?> from, Class<?> fromClass) throws Exception {

		if (SortedMap.class.isAssignableFrom(fromClass)) {
			SortedMap<?, ?> fromSortedMap = (SortedMap<?, ?>) from;
			Comparator<? super Object> toComparator = newComparator(fromSortedMap);
			return new TreeMap<Object, Object>(toComparator);
		}
		if (isJavaOrAppPackage(fromClass) && !fromClass.isInterface()) {
			if (SortedMap.class.isAssignableFrom(fromClass)) {
				SortedMap<?, ?> fromSortedMap = (SortedMap<?, ?>) from;
				Comparator<?> toComparator = newComparator(fromSortedMap);

				if (toComparator != null)
					return newInstanceWithComparator(fromSortedMap.getClass(),
							toComparator);
			}
			return (Map<Object, Object>) newInstance(fromClass);
		}
		return new HashMap<Object, Object>();
	}

	/**
	 * Creates a non cglib/javassist enhanced instance of the given class, which
	 * could itself be the class of a cglib/javassist enhanced object.
	 */
	protected Object newInstance(Class<?> fromClass) throws Exception {
		Class<?> toClass = fromClass;

		if (ClassUtil.isEnhanced(toClass)) {
			// figure out the pre-enhanced class
			toClass = toClass.getSuperclass();
		}

		try {
			return newInstanceAsPrivileged(toClass);
		} catch (Exception e) {
			throw ClassUtil.wrapRun(e);
		}	
	}

	@SuppressWarnings("unchecked")
	private <T> T newInstanceWithComparator(Class<T> fromClass,
			Comparator<?> comparator) throws Exception {
		Class<?> toClass = fromClass;

		if (ClassUtil.isEnhanced(toClass)) {
			// figure out the pre-enhanced class
			toClass = toClass.getSuperclass();
		}

		return (T) newInstanceWithComparatorAsPrivileged(toClass, comparator);
	}
	
	private static Constructor getConstructor(Class<?> c) {
		Constructor[] ctors = c.getDeclaredConstructors();
		Constructor ctor = null;
		for (int i = 0; i < ctors.length; i++) {
		    ctor = ctors[i];
		    if (ctor.getGenericParameterTypes().length == 0)
			break;
		}		
		
		return ctor;
	}

	/**
	 * Creates a new instance of the given class via the no-arg constructor,
	 * invoking the constructor as a privileged action if it is protected or
	 * private.
	 * 
	 * @param c
	 *            given class
	 * @return a new instance of the given class via the no-arg constructor
	 */
	public static Object newInstanceAsPrivileged(Class<?> c) throws Exception {
		final Constructor<?> constructor = c.getDeclaredConstructor();

		if (Modifier.isPublic(constructor.getModifiers()))
			return c.newInstance();

		return AccessController.doPrivileged(new PrivilegedAction<Object>() {
			public Object run() {
				constructor.setAccessible(true);
				try {
					return constructor.newInstance();
				} catch (Exception e)  {
					throw ClassUtil.wrapRun(e);
				}
			}
		});
	}

	private Object newInstanceWithComparatorAsPrivileged(Class<?> c,
			final Comparator<?> comparator) throws Exception {
		final Constructor<?> constructor = c
				.getDeclaredConstructor(Comparator.class);

		if (Modifier.isPublic(constructor.getModifiers()))
			return constructor.newInstance(comparator);

		return AccessController.doPrivileged(new PrivilegedAction<Object>() {
			public Object run() {
				constructor.setAccessible(true);
				try {
					return constructor.newInstance(comparator);
				} catch (Exception e)  {
					throw ClassUtil.wrapRun(e);
				}				
			}
		});
	}

	public void clearVisited() {
		for(Object object: sourceTargetMap.values()) {
			Node node = (Node) object;
			node.setVisited(false);
		}
	}	
}
